//--------------------------------------------------------------------------
///	\file	Light.cpp
/// \author	Rob Bateman			( robthebloke@hotmail.com )
/// \brief	This file defines the single function OutputLight which will write the light
///			data to the file. 
/// \note       
/// 
/// 		This sourcefile is provided as-is and no guarentee or warrenty 
///			is provided for any damage that may arise from using it. This
///			code is free for your own use, the only thing i ask is a small
///			credit somewhere for my work. An e-mail saying you found it 
///			useful would also be much appreciated by myself ;)
///
///			If you find any bugs arising to maya version changes or you
///			have improvements to the source code, then feel free to mail
///			them to me and i'll update the source-code on the website.
///			Before you e-mail bug reports, make sure that it is my code 
///			causing the problem, and not some wierd oddity in the maya 
///			scene. This code is designed to handle most common model set-ups,
///			it is not designed to replicate perfectly everything that maya
///			does. 
/// 
///			DO NOT send me feature requests, i'm not likely to add to this
///			code myself unless something really new and groovy comes along.
///			This is mainly due to my current time restrictions. If you do
///			have code written that adds something fairly groovy to the 
///			exporter, then i will happily update the exporter code and give
///			full credit where due.....
/// 
///			Robert Bateman
/// 
///				email code to:    rbateman@bmth.ac.uk
///				contact:          robthebloke@hotmail.com
///	
//--------------------------------------------------------------------------


#ifdef WIN32
	#ifndef NT_APP
		#define NT_APP
	#endif

	// disable VC6 template name warning
	#pragma warning(disable:4786)
#endif


#include "Light.h"
#include "Mesh.h"

#include <maya/MColor.h>
#include <maya/MMatrix.h>
#include <maya/MVector.h>

#include <maya/MFnAmbientLight.h>
#include <maya/MFnAreaLight.h>
#include <maya/MFnDagNode.h>
#include <maya/MFnDirectionalLight.h>
#include <maya/MFnLight.h>
#include <maya/MFnPointLight.h>
#include <maya/MFnSpotLight.h>
#include <maya/MFnTransform.h>

#include <iostream>
#include <cassert>

/////////////////////////////////////////////////////////////////////////////////////////
/// \brief	This function outputs light data to the specifed filestream. The MObject
///			passed as an argument is the handle to the maya node.
/// \param	ofs	-	the output stream
///	\param	obj	-	the maya object
/// 
void OutputLight(std::ofstream& ofs,MObject &obj)
{
	// make sure we only get valid nodes here
	assert( !obj.isNull() );

	// make sure the MObject represents a light!
	assert( obj.hasFn(MFn::kLight) );

	// attach a light function set to the object
	MFnLight fnLight(obj);

	// if the object is an intermediate object, we will ignore it
	if ( !fnLight.isIntermediateObject() ) 
	{
		// get light color
		MColor color(0.0f,0.0f,0.0f);
		color = fnLight.color();

		// point at zero
		MVector p(0,0,0);

		// transform it by the light x-form matrix
		//
		MFnTransform fnXform( fnLight.parent(0) );

		// get position of the light
		//
		MMatrix ws = BuildTransform(fnXform.object());

		// output some light info
		ofs << "LIGHT "

			// output name
			<< fnLight.name().asChar()
			
			// output light type
			<< "\n\tType=\t" 
			<< obj.apiTypeStr()

			// output color 
			<< "\n\tColor=\t"
			<< color.r
			<< " "
			<< color.g
			<< " "
			<< color.b 

			// output intensity
			<< "\n\tIntensity=\t" 
			<< fnLight.intensity()

			// output light position
			<< "\n\tPosition=\t"
			<< ws(3,0) << " " << ws(3,1) << " " << ws(3,2)
			<< std::endl;


		// if you need to get specific attributes associated with specific
		// light types then use something like this....
		//
		switch(obj.apiType()) 
		{
		case MFn::kPointLight:
			{
				MFnPointLight fnPoint(obj);

			}
			break;
		case MFn::kAmbientLight:
			{
				MFnAmbientLight fnAmbient(obj);

			}
			break;
		case MFn::kSpotLight:
			{
				MFnSpotLight fnSpot(obj);

			}
			break;
		case MFn::kDirectionalLight:
			{
				MFnDirectionalLight fnDir(obj);

			}
			break;
		case MFn::kAreaLight:
			{
				MFnAreaLight fnArea(obj);

			}
			break;
		default:
			break;
		}
	}
}
