/////////////////////////////////////////////////////////////////////////////////////////
///
///	\file	Material.cpp
/// \author	Rob Bateman
/// \brief	This file defines the single function OutputMaterials which will output all
///			of the material information for phong,lambert and blinn shaders to the
///			output files. 
/// \note       
/// 
/// 		This sourcefile is provided as-is and no guarentee or warrenty 
///			is provided for any damage that may arise from using it. This
///			code is free for your own use, the only thing i ask is a small
///			credit somewhere for my work. An e-mail saying you found it 
///			useful would also be much appreciated by myself ;)
///
///			If you find any bugs arising to maya version changes or you
///			have improvements to the source code, then feel free to mail
///			them to me and i'll update the source-code on the website.
///			Before you e-mail bug reports, make sure that it is my code 
///			causing the problem, and not some wierd oddity in the maya 
///			scene. This code is designed to handle most common model set-ups,
///			it is not designed to replicate perfectly everything that maya
///			does. 
/// 
///			DO NOT send me feature requests, i'm not likely to add to this
///			code myself unless something really new and groovy comes along.
///			This is mainly due to my current time restrictions. If you do
///			have code written that adds something fairly groovy to the 
///			exporter, then i will happily update the exporter code and give
///			full credit where due.....
/// 
///			Robert Bateman
/// 
///				email code to:    rbateman@bmth.ac.uk
///				contact:          robthebloke@hotmail.com
///	
//--------------------------------------------------------------------------


#ifdef WIN32
	#ifndef NT_APP
		#define NT_APP
	#endif

	// disable VC6 template name warning
	#pragma warning(disable:4786)
#endif


#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <map>

#include <maya/MItDependencyNodes.h>
#include <maya/MFnDependencyNode.h>
#include <maya/MFnPhongShader.h>
#include <maya/MFnLambertShader.h>
#include <maya/MFnBlinnShader.h>
#include <maya/MPlug.h>
#include <maya/MPlugArray.h>
#include <maya/MColor.h>

#include "Material.h"
#include "Texture.h"



// this will hold the name of the kFileTexture node and an ID number we will
// create to reference the textures from the model file.
//
std::map<std::string,unsigned int> g_MaterialToID_Map; 
unsigned int mat_id_gen = 0;


unsigned int GetMaterialID(const MString &material) {
	std::map<std::string,unsigned int>::iterator it = g_MaterialToID_Map.find(material.asChar());

	if (it != g_MaterialToID_Map.end()) {
		return it->second;
	}
	return 0;
}


/// \brief	This function outputs the material color via manually getting a plug to the attribute
///			this is just to make things a bit more generic.
/// \param	ofs		-	the output file stream
/// \param	fn		-	the function set
/// \param	name	-	name of the attribute
/// 
bool OutputMaterialColor(std::ostream& ofs,MFnDependencyNode& fn,std::string name)
{
	MPlug p;
	
	std::string r = name;
	r += "R";
	std::string g = name;
	g += "G";
	std::string b = name;
	b += "B";
	std::string a = name;
	a += "A";

	// get the color value
	MColor color;
	
	// get a plug to the attribute
	p = fn.findPlug(r.c_str());
	p.getValue(color.r);
	p = fn.findPlug(g.c_str());
	p.getValue(color.g);
	p = fn.findPlug(b.c_str());
	p.getValue(color.b);
	p = fn.findPlug(a.c_str());
	p.getValue(color.a);

	p = fn.findPlug(name.c_str());

	// will hold the txture node name
	MString texname;

	// get plugs connected to colour attribute
	MPlugArray plugs;
	p.connectedTo(plugs,true,true);

	// see if any file textures are present
	for(int i=0;i!=plugs.length();++i)
	{
		// if file texture found
		if (plugs[i].node().apiType() == MFn::kFileTexture) 
		{
			// bind a function set to it ....
			MFnDependencyNode fnDep(plugs[i].node());

			// to get the node name
			texname = fnDep.name();

			// stop looping
			break;
		}
	}

	if (name == "color") {
		color.r=color.g=color.b=0.6f;
	}

	// output the name, color and texture ID
	ofs << "\t" << name << " "
		<< color.r << " "
		<< color.g << " "
		<< color.b << " "
		<< color.a << " tex= "
		<< GetTextureID( texname ) << "\n";

	return true;
}

/// \brief	This function traverses the scene and writes out the material data. At the moment
///			I am just supporting three types Lambert, Phong and Blinn. 
/// \param	
void OutputMaterials(std::ostream &ofs)
{
	// iterate through the mesh nodes in the scene
	//
	MItDependencyNodes itDep(MFn::kLambert);

	// we have to keep iterating until we get through all of the nodes in the scene
	//
	while (!itDep.isDone()) 
	{

		switch(itDep.item().apiType()) {

			// if found phong shader
			case MFn::kPhong:
				{
					MFnPhongShader fnPhong( itDep.item() );

					g_MaterialToID_Map.insert( std::make_pair( fnPhong.name().asChar(), ++mat_id_gen ) );

					ofs << "PHONG " << fnPhong.name().asChar() << "\n";
					OutputMaterialColor(ofs,fnPhong,"ambientColor");
					OutputMaterialColor(ofs,fnPhong,"color");
					OutputMaterialColor(ofs,fnPhong,"specularColor");
					OutputMaterialColor(ofs,fnPhong,"incandescence");
					OutputMaterialColor(ofs,fnPhong,"transparency");
					ofs	<< "\tshininess " << (fnPhong.cosPower()) << std::endl; 

				}
				break;

			// if found lambert shader
			case MFn::kLambert:
				{
					MFnLambertShader fnLambert( itDep.item() );

					g_MaterialToID_Map.insert( std::make_pair( fnLambert.name().asChar(), ++mat_id_gen ) );

					ofs << "LAMBERT " << fnLambert.name().asChar() << "\n";
					OutputMaterialColor(ofs,fnLambert,"ambientColor");
					OutputMaterialColor(ofs,fnLambert,"color");
					OutputMaterialColor(ofs,fnLambert,"incandescence");
					OutputMaterialColor(ofs,fnLambert,"transparency");
				}
				break;

			// if found blinn shader
			case MFn::kBlinn:
				{
					MFnBlinnShader fnBlinn( itDep.item() );

					g_MaterialToID_Map.insert( std::make_pair( fnBlinn.name().asChar(), ++mat_id_gen ) );

					ofs << "BLINN " << fnBlinn.name().asChar() << "\n";
					OutputMaterialColor(ofs,fnBlinn,"ambientColor");
					OutputMaterialColor(ofs,fnBlinn,"color");
					OutputMaterialColor(ofs,fnBlinn,"specularColor");
					OutputMaterialColor(ofs,fnBlinn,"incandescence");
					OutputMaterialColor(ofs,fnBlinn,"transparency");
					ofs	<< "\teccentricity " << (fnBlinn.eccentricity()) << std::endl; 
					ofs	<< "\tspecularRollOff " << (fnBlinn.specularRollOff()) << std::endl; 
				}
				break;
			default:
				break;
		}

		itDep.next();
	}
}

