//--------------------------------------------------------------------------
///	\file	Mesh.cpp
/// \author	Rob Bateman
/// \brief	This file defines the single function OutputMesh which will write the mesh
///			data to the file.
/// \note	When accessing any node data within maya, you must use a function set
///			and attach it to the relevant MObject within maya. All function sets
///			are listed with the prefix MFn. The class MFn also contains the type
///			enumerations for the various maya nodes, in the case of a mesh,
///
///									MFn::kMesh
///
///
/// 		This sourcefile is provided as-is and no guarentee or warrenty
///			is provided for any damage that may arise from using it. This
///			code is free for your own use, the only thing i ask is a small
///			credit somewhere for my work. An e-mail saying you found it
///			useful would also be much appreciated by myself ;)
///
///			If you find any bugs arising to maya version changes or you
///			have improvements to the source code, then feel free to mail
///			them to me and i'll update the source-code on the website.
///			Before you e-mail bug reports, make sure that it is my code
///			causing the problem, and not some wierd oddity in the maya
///			scene. This code is designed to handle most common model set-ups,
///			it is not designed to replicate perfectly everything that maya
///			does.
///
///			DO NOT send me feature requests, i'm not likely to add to this
///			code myself unless something really new and groovy comes along.
///			This is mainly due to my current time restrictions. If you do
///			have code written that adds something fairly groovy to the
///			exporter, then i will happily update the exporter code and give
///			full credit where due.....
///
///			Robert Bateman
///
///				email code to:    rbateman@bmth.ac.uk
///				contact:          robthebloke@hotmail.com
///
//--------------------------------------------------------------------------

#ifdef WIN32
	#ifndef NT_APP
		#define NT_APP
	#endif

	// disable VC6 template name warning
	#pragma warning(disable:4786)
#endif


//--------------------------------------------------------------------------
// includes
//
#include <maya/MFloatMatrix.h>
#include <maya/MFloatVector.h>
#include <maya/MFloatVectorArray.h>
#include <maya/MIntArray.h>
#include <maya/MMatrix.h>
#include <maya/MObjectArray.h>
#include <maya/MPlug.h>
#include <maya/MPlugArray.h>
#include <maya/MPoint.h>
#include <maya/MPointArray.h>
#include <maya/MObjectArray.h>

#include <maya/MFnDependencyNode.h>
#include <maya/MFnMesh.h>
#include <maya/MFnTransform.h>

#include <maya/MItDependencyNodes.h>
#include <maya/MItMeshPolygon.h>
#include <maya/MItMeshVertex.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <map>


#include "Mesh.h"
#include "Material.h"


MMatrix BuildTransform(const MObject& obj)
{
	MFnTransform fn(obj);

	// get matrix transform for this bone
	MMatrix M = fn.transformationMatrix();

	if (fn.parentCount()>0) {
		return M* BuildTransform(fn.parent(0)) ;
	}
	return M;
}



//--------------------------------------------------------------------------
/// \brief	This function outputs the specified mesh to the specified file stream
///
void OutputMesh(std::ofstream& ofs,const MObject &obj)
{
	// attach a function set to the mesh node. We access all of the meshes data
	// through the function set
	//
	MFnMesh fnMesh(obj);

	// Maya has a tendancy of storing additional nodes and objects all over the place.
	// We may come accross meshes that are nodes used purely as history or other
	// construction data. In our case, we want to ignore all 'intermediate objects'
	//
	if ( !fnMesh.isIntermediateObject() )
	{
		for(int k=0;k<fnMesh.parentCount();++k)
		{
			int i,j;

			// write out some info about the mesh
			//
			ofs << "MESH "
				<< fnMesh.name().asChar()
				<< "\n\tVertexCount= "
				<< fnMesh.numVertices()
				<< "\n\tNormalCount= "
				<< fnMesh.numNormals()
				<< "\n\tTexCoordCount= "
				<< fnMesh.numUVs()
				<< "\n\tFaceCount= "
				<< fnMesh.numPolygons()
				<< std::endl;


			// this will hold the returned vertex positions
			MPointArray vertices;

			// use the function set to get the points
			fnMesh.getPoints(vertices);



			// Here we just grab the transform matrix to put the mesh into world spaces
			//
			MMatrix mat = BuildTransform( fnMesh.parent(k) );


			// go through each vertex in the array and write it to the file
			//
			ofs << "\n\tVertices\n";
			for(i=0;i!=vertices.length() ; ++i)
			{
				// transform the points from local space to world space
				//
				MPoint OutPoint = vertices[i] * mat;

				// output
				//
				ofs << "\t\t" << OutPoint.x << " " << OutPoint.y << " " << OutPoint.z << "\n";
			}



			// this will store the normals
			MFloatVectorArray normals;

			// get the normals
			fnMesh.getNormals(normals);

			// just to be an utter arse, normals require you to have float matrices.
			// Unfortunately someone decided not to allow you to convert from the
			// MMatrix class to the MFloatMatrix class, therefore we have to do it
			// in a very very gay way...
			//
			MFloatMatrix fmat;

			for(i=0;i!=4;++i)
			{
				for(j=0;j!=4;++j)
				{
					fmat.matrix[i][j] = static_cast<float>(mat.matrix[i][j]);
				}
			}

			// in order to transform normals, we need to multiply them by the inverse transpose
			// of the world space transform matrix. Might as well invert & transpose the matrix then....
			fmat = fmat.inverse();
			fmat = fmat.transpose();

			// go through each normal in the array and write it to the file
			//
			ofs << "\n\tNormals\n";
			for(i=0;i!=normals.length() ; ++i)
			{

				// transform the normal into world space
				//
				MFloatVector OutNormal = normals[i] * fmat;

				// we must normalise the vector to counter any scaling in the transformation
				//
				OutNormal.normalize();

				// output the normal
				//
				ofs << "\t\t" << OutNormal.x << " " << OutNormal.y << " " << OutNormal.z << "\n";
			}

			if(fnMesh.numUVs()>0)
			{

				// these two arrays will hold the u & v texture co-ordinates
				//
				MFloatArray u_coords;
				MFloatArray v_coords;

				// get the uv coordinates
				//
				ofs << "\n\tTexCoords\n";
				fnMesh.getUVs(u_coords,v_coords);

				// go through each uv in the array and write it to the file
				//
				for(i=0;i!=u_coords.length() ; ++i)
				{
					ofs << "\t\t" << u_coords[i] << " " << v_coords[i] << "\n";
				}
			}


			// this will hold references to the shaders used on the meshes
			MObjectArray Shaders;

			// this is used to hold indices to the materials returned in the object array
			MIntArray    FaceIndices;


			// this function builds a list of all surface shaders used on the mesh. Note
			// that the nodes returned are the shading groups, not the material nodes!
			fnMesh.getConnectedShaders(k,Shaders,FaceIndices);

			// having recieved the shader information, we will try to process the returned
			// shaders to find their name and ultimately find the relevant material ID that
			// references the list we have built up.
			std::vector< unsigned int > ShaderIDs;

			// start the first shader as a default zero value
			ShaderIDs.push_back(0);


			int kk = 0;

			// for each surface shader
			for(kk=0;kk!=Shaders.length();++kk)
			{
				// attachy function set...
				MFnDependencyNode fnShader(Shaders[kk]);

				// get access to the surfaceShader attribute. This will be connected to
				// lambert , phong nodes etc.
				MPlug sshader = fnShader.findPlug("surfaceShader");

				MPlugArray materials;

				// get the material connected to the surface shader
				sshader.connectedTo(materials,true,true);

				//
				for(unsigned int l=0;l<materials.length();++l)
				{
					MFnDependencyNode fnMat(materials[l].node());
					ShaderIDs.push_back( GetMaterialID(fnMat.name()) );
				}
			}



			kk = 0;


			// to get at the face data, we need to use a mesh polygon iterator.
			//
			MItMeshPolygon itPoly(fnMesh.object());

			// keep looping until the iterator has finished
			//
			while ( !itPoly.isDone() )
			{

				// get the number of vertices in this face
				unsigned int vcount = itPoly.polygonVertexCount();
				int matInd;
					{
						matInd =  ShaderIDs[ FaceIndices[kk]+1 ];
					}
				// output some info about the face
				//
				ofs << "\n\tFACE "
					<< vcount
					<< "\n\t\tMat= "
					<< matInd
					<< std::endl;

				// for each vertex on the face....
				//
				for(i = 0; i!= vcount; ++i)
				{


					// output indices for this vertex
					ofs << "\t\t" << itPoly.vertexIndex(i) << " " << itPoly.normalIndex(i) ;
					if(fnMesh.numUVs()>0)
					{
						// have to get the uv index seperately
						int uv_index;
						itPoly.getUVIndex(i,uv_index);
						ofs << " " << uv_index;
					}
					ofs << "\n";
				}

				++kk;

				// move to next poly in the mesh
				//
				itPoly.next();
			}
			ofs << "\n";
		}
	}
}
