//---------------------------------------------------------------------------------------
/// 
///	\file	Texture.cpp
/// \author	Rob Bateman			( robthebloke@hotmail.com )
/// \brief	OK, now we are going to look at textures (note - not materials). Maya offers 
///			us the really handy class MImage which allows us to load (though not save) 
///			any file format supported by maya. This basically allows us to traverse the 
///			scene looking for file texture nodes, read the filename, load the file and 
///			then output the data in our own format. This may seem a bit long winded, 
///			however it basically gives us a HELL of a lot  of freedom because we don't 
///			need to know how to load hectic file formats such as jpg tiff etc. It also 
///			gives us an opportunity to check the texture size. 
///
///			OpenGL only supports textures that have dimension that is of power 2, ie 
///			2,4,8,16,32 etc. Now, if the texture file is not power of 2, we can simply 
///			resize it to a valid size and print a warning. The constant g_DefaultTextureRes 
///			exists for this purpose. Currently this is set to 128. You could be a bit 
////		cleverer here by introducing some heuristic to choose what size to scale the 
///			texture to. 
///			
///			When it comes to storing this texture data we really don't want to store it 
///			in the same file as the model data, mainly because there is FUCK loads of 
///			texture data, so it's better to package all of the textures together into a
///			single file. Basically to ensure that we never loose textures or have problems
///			with Full paths names when transferred to another computer, we will simply use 
///			an ID referencing system that will refer to a one based 
///			index which represents the specific texture file. (0 means no texture)
///			
///			For each image we want to store :
///			
///			width	
///			height	
///			bpp		1/3/4
///			
///			It's index is given by it's place in the file (starting at one).
///			
///			Now we could hold that data as 3 shorts, or 2 shorts and a bool. Thats a bit cack. 
///			I kinda like nice small fields, and a bit of data compaction. Instead of storing the
///			width in pixels, why not just store the power? With 4bits of data (a nibble) we can 
///			store a value between 0 and 15, ie 2^0 to 2^15, or 1 to 32768. So in 1byte we can store
///			the width and the height! If we then add 1byte to store the bpp we should have minimal
///			data use and a couple of bits left over (possibly compressed data or colour table flags?)	
/// \note       
/// 
/// 		This sourcefile is provided as-is and no guarentee or warrenty 
///			is provided for any damage that may arise from using it. This
///			code is free for your own use, the only thing i ask is a small
///			credit somewhere for my work. An e-mail saying you found it 
///			useful would also be much appreciated by myself ;)
///
///			If you find any bugs arising to maya version changes or you
///			have improvements to the source code, then feel free to mail
///			them to me and i'll update the source-code on the website.
///			Before you e-mail bug reports, make sure that it is my code 
///			causing the problem, and not some wierd oddity in the maya 
///			scene. This code is designed to handle most common model set-ups,
///			it is not designed to replicate perfectly everything that maya
///			does. 
/// 
///			DO NOT send me feature requests, i'm not likely to add to this
///			code myself unless something really new and groovy comes along.
///			This is mainly due to my current time restrictions. If you do
///			have code written that adds something fairly groovy to the 
///			exporter, then i will happily update the exporter code and give
///			full credit where due.....
/// 
///			Robert Bateman
/// 
///				email code to:    rbateman@bmth.ac.uk
///				contact:          robthebloke@hotmail.com
///	
//---------------------------------------------------------------------------------------

#ifdef WIN32
	#ifndef NT_APP
		#define NT_APP
	#endif

	// disable VC6 template name warning
	#pragma warning(disable:4786)
#endif



//---------------------------------------------------------------------------------------

#include "Texture.h"

#include <maya/MImage.h>
#include <maya/MPlug.h>
#include <maya/MString.h>
#include <maya/MObjectArray.h>

#include <maya/MFnDependencyNode.h>
#include <maya/MItDependencyNodes.h>

#include <iostream>
#include <string>
#include <map>


//---------------------------------------------------------------------------------------


// if the texture is not of power 2, then it will be resized to this
// default resolution.
// 
const unsigned int g_DefaultTextureRes = 256;

/// this will hold the name of the kFileTexture node and an ID number we will
/// create to reference the textures from the model file.
///
std::map<std::string,unsigned int> g_TextureToID_Map; 

/// This variable is used to generate a unique ID for the texture nodes. These
/// ID's are the ones used within the above map
///
unsigned int id_gen=0;



//---------------------------------------------------------------------------------------
/// \brief	This function is another shady work around. The one problem with MImage is
///			that it always converts image data to RGBA. This is a bit of a pain because 
///			Ideally we would like a greyscale / 24bit and 32bit options. You may be able 
///			to hack around the problem with naming conventions, however I'm being very 
///			lazy and am checking every alpha value to see if it is the same as the first.
///			If they are all the same then it has a very boring alpha channel that I don't
///			want in my game enigne :p
///
bool Is24Bit(const unsigned char* px,unsigned short w, unsigned short h)
{
	unsigned char first = px[3];
	for(unsigned int i=7;i<w*h*4;i+=4)
	{
		if(first != px[i])
		{
			return false;
		}
	}
	return true;
}

//---------------------------------------------------------------------------------------
/// \brief	Is i a power of two var....
/// 
bool IsPowerOfTwo(unsigned int i)
{
	if( i == 2   || i == 4   ||
		i == 8   || i == 16  ||
		i == 32  || i == 64  ||
		i == 128 || i == 256 ||
		i == 512 || i == 1024 ) 
	{
			return true;
	}
	return false;
}

//---------------------------------------------------------------------------------------
/// \brief	Takes the value v and returns the power
/// 
unsigned char LogIt(unsigned short v)
{
	switch(v) {
	case 1:
		return 0;
	case 2:
		return 1;
	case 4:
		return 2;
	case 8:
		return 3;
	case 16:
		return 4;
	case 32:
		return 5;
	case 64:
		return 6;
	case 128:
		return 7;
	case 256:
		return 8;
	case 512:
		return 9;
	case 1024:
		return 10;
	case 2048:
		return 11;
	case 4096:
		return 12;
	case 8192:
		return 13;
	case 16384:
		return 14;
	case 32768:
		return 15;
	default:
		break;
	}
	return 0;
}


//---------------------------------------------------------------------------------------
/// \brief	Function to save a texture to the specified file pointer
/// \param	fp	-	the file pointer 
///	\param	img	-	reference to the MImage class with the loaded data
///
void OutputTexture(FILE *fp,MImage &img)
{
	ImageHeader header;
	unsigned int w,h;

	// get image size of the texture
	img.getSize(w,h);

	bool bShouldResize = false;

	// check dimesnions
	if(!IsPowerOfTwo(w))
	{
		bShouldResize = true;
		w = g_DefaultTextureRes;	
	}
	if(!IsPowerOfTwo(h)) 
	{
		bShouldResize = true;
		h = g_DefaultTextureRes;
	}

	if (bShouldResize) 
	{
		std::cerr << "Warning image not power of 2. Resizing to (" 
				  << w 
				  << "," 
				  << h 
				  << ")" 
				  << std::endl;

		// oh dear, resize the image 
		img.resize(w,h,false);

	}

	// store width and height in the header
	header.m_WidthPower  = LogIt(w);
	header.m_HeightPower = LogIt(h);

	// store pixel depth in header
	if (Is24Bit(img.pixels(),w,h)) {
		header.m_Bpp		 = 2;
	}
	else {
		header.m_Bpp		 = 3;
	}

	// write header to the file
	fwrite(&header,sizeof(ImageHeader),1,fp);

	// write the 32 bit data to the file
	if (header.m_Bpp == 3) {
		fwrite(img.pixels(),1,w*h*4,fp);
	}
	else
	{
		// write 24bit data in chunks so we skip the fourth alpha byte
		for(unsigned int i=0;i<w*h*4;i+=4)	{
			fwrite(img.pixels() + i,1,3,fp);
		}
	}
}





//---------------------------------------------------------------------------------------
/// \brief	This function returns the texture ID of the texture node specified.
///			If the name specified does not exist it will return 0.
/// 
unsigned int GetTextureID(MString &fnn) {

	// try and find the specified texture ID
	std::map<std::string,unsigned int>::iterator it = g_TextureToID_Map.find(fnn.asChar());

	// if found ...
	if (it != g_TextureToID_Map.end()) {

		// return the ID
		return it->second;
	}

	// return 0 for invalid texture ID
	return 0;
}


//---------------------------------------------------------------------------------------
/// \brief	This function outputs all of the texture data within a scene and writes
///			it as a single file. The file starts with a 2byte short giving the 
///			number of image files present. Each image has a 2 byte ImageHeader struct
///			which details the width, height and pixel depth of the following image. 
///			The image data is uncompressed, RGB,RGBA data.
/// 
void DoTextures(const char * const outfile)
{
	// vector to hold the texture filenames
	MObjectArray FileNodes;

	// create an iterator for the texture files
	MItDependencyNodes itDep(MFn::kFileTexture);

	// set the id generator to zero
	id_gen = 0;

	// loop through fileTexture nodes until done
	while (!itDep.isDone()) 
	{
		// create a function set to get some useful attributes
		MFnDependencyNode fnDep(itDep.item());

		// store the MObject to the file texture
		FileNodes.append(itDep.item());

		// insert the name of the node and the ID number into the map for lookup when
		// exporting materials.
		g_TextureToID_Map.insert( std::make_pair( fnDep.name().asChar(), ++id_gen) );

		// move to next texture node
		//
		itDep.next();
	}

	// open the output file
	FILE* fp = fopen(outfile,"wb");

	// if OK
	if (fp) {

		// the first 2 bytes of the file is a number identifying how many textures
		// are in this file. 
		unsigned short texCount = FileNodes.length();
		fwrite(&texCount,sizeof(unsigned short),1,fp);

		// iterate through all FileNodes
		for( int i=0 ; i != FileNodes.length(); ++i )
		{
			MImage img;
		
			// attempt to read the image file
			if( MS::kSuccess == img.readFromTextureNode( FileNodes[i] ) )
			{
				// output the texture if sucessful
				OutputTexture(fp,img);
				img.release();
			}
			else
			{
				// output warning and 2byte field of zeros if not opened
				unsigned short temp = 0x0000;
				fwrite(&temp,sizeof(unsigned short),1,fp);
			}

		}

		// close the file when done.
		fclose(fp);
	}

}
