//-------------------------------------------------------------------------- 
///	\file	Texture.h
/// \author	Rob Bateman			( robthebloke@hotmail.com )
/// \brief	This file defines a bunch of functions to grab and export texture 
///			data. The DoTextures function should be called before reading back 
///			any material data. This will write all textures to a single binary 
///			file. The textures are then assigned a 1 based index ID number. This 
///			information is stored internally in a map that uses the texture node 
///			name (NOT THE FILENAME!) as the key and the ID as the value. 
/// 
///			This then allows you to use the GetTextureID() function later on to 
///			find the reference number of the texture file.
/// 
///			CleanupMyTextures() should be called after the file has been 
///			exported. If not, residual data may cause problems (not an 
///			issue just yet....)
/// \note       
/// 
/// 		This sourcefile is provided as-is and no guarentee or warrenty 
///			is provided for any damage that may arise from using it. This
///			code is free for your own use, the only thing i ask is a small
///			credit somewhere for my work. An e-mail saying you found it 
///			useful would also be much appreciated by myself ;)
///
///			If you find any bugs arising to maya version changes or you
///			have improvements to the source code, then feel free to mail
///			them to me and i'll update the source-code on the website.
///			Before you e-mail bug reports, make sure that it is my code 
///			causing the problem, and not some wierd oddity in the maya 
///			scene. This code is designed to handle most common model set-ups,
///			it is not designed to replicate perfectly everything that maya
///			does. 
/// 
///			DO NOT send me feature requests, i'm not likely to add to this
///			code myself unless something really new and groovy comes along.
///			This is mainly due to my current time restrictions. If you do
///			have code written that adds something fairly groovy to the 
///			exporter, then i will happily update the exporter code and give
///			full credit where due.....
/// 
///			Robert Bateman
/// 
///				email code to:    rbateman@bmth.ac.uk
///				contact:          robthebloke@hotmail.com
///	
//--------------------------------------------------------------------------

#ifdef WIN32
#pragma once
#endif

#ifndef TEXTURE_H_INCLUDED
#define TEXTURE_H_INCLUDED

#include <maya/MString.h>


//--------------------------------------------------------------------------
/// \brief	This struct is used as a 2byte identifier just before an image
/// \note	sizeof(ImageHeader) == 2
/// 
struct ImageHeader
{
	/// the width of the image as the power of two
	unsigned char m_WidthPower:4;
	/// the height of the image as a power of two
	unsigned char m_HeightPower:4;

	/// the bits per pixel
	/// 0 = 8bit
	/// 1 = 16bit
	/// 2 = 24bit
	/// 3 = 32bit
	/// 
	unsigned char m_Bpp:2;

	/// 6bits for a rainy day...
	unsigned char m_Unused:6;

	ImageHeader() : m_WidthPower(0),m_HeightPower(0),m_Bpp(0),m_Unused(0){

	}
};



//--------------------------------------------------------------------------
/// \brief	This function outputs all of the textures in the scene into a single 
///			binary file. 
/// \param	outfile	-	the name of the binary texture file to write
/// 
void DoTextures(const char *const);


//--------------------------------------------------------------------------
/// \brief	This function returns the texture ID of the texture node specified.
///			If the name specified does not exist it will return 0.
/// \param	fnn	-	the file node name. NOTE This is not the name of the texture
///			file, but the name of the fileTexture node in maya.
/// 
unsigned int GetTextureID(MString&);


//--------------------------------------------------------------------------
/// \brief	This function cleans all dynamically allocated data (just the map infact)
/// 
void CleanupMyTextures();

#endif
