//--------------------------------------------------------------------------
/// 
/// \file	Material.h
/// \author	Rob Bateman
/// \brief	I've tried to make this loader a bit more generic. The material 
///			class is now able to store colour channels, ie a colour and a texture
///			for ambient, diffuse, specular etc. With basic openGL this isn't of 
///			much use because you can only use 1 texture 
///			(though 2 with mutlitexturing extension). This is mainly done for my
///			own purposes cos I use CG shaders so a bit more flexibility is nice.
/// 
///			At the moment, if a texture is assigned to the diffuse channel, then 
///			it is used as the surface texture.
/// \note       
/// 
/// 		This sourcefile is provided as-is and no guarentee or warrenty 
///			is provided for any damage that may arise from using it. This
///			code is free for your own use, the only thing i ask is a small
///			credit somewhere for my work. An e-mail saying you found it 
///			useful would also be much appreciated by myself ;)
///
///			If you find any bugs arising to maya version changes or you
///			have improvements to the source code, then feel free to mail
///			them to me and i'll update the source-code on the website.
///			Before you e-mail bug reports, make sure that it is my code 
///			causing the problem, and not some wierd oddity in the maya 
///			scene. This code is designed to handle most common model set-ups,
///			it is not designed to replicate perfectly everything that maya
///			does. 
/// 
///			DO NOT send me feature requests, i'm not likely to add to this
///			code myself unless something really new and groovy comes along.
///			This is mainly due to my current time restrictions. If you do
///			have code written that adds something fairly groovy to the 
///			exporter, then i will happily update the exporter code and give
///			full credit where due.....
/// 
///			Robert Bateman
/// 
///				email code to:    rbateman@bmth.ac.uk
///				contact:          robthebloke@hotmail.com
///	
//--------------------------------------------------------------------------

#ifdef WIN32
#pragma once
#endif

#ifndef __X_MATERIAL__H_
#define __X_MATERIAL__H_

#include <vector>
#include <iostream>

//--------------------------------------------------------------------------
class XMaterial
{
public:
	
	XMaterial(void);
	~XMaterial(void);

	/// the function to bind the material
	void gl();

	/// a function to set the ambient colour
	inline void SetAmbient(const float r,
						   const float g,
						   const float b,
						   const float a=1.0f) {
		m_Ambient.SetColour(r,g,b,a);
	}

	/// a function to set the diffuse colour
	inline void SetDiffuse(const float r,
						   const float g,
						   const float b,
						   const float a=1.0f) {
		m_Diffuse.SetColour(r,g,b,a);
	}

	/// a function to set the specular colour
	inline void SetSpecular(const float r,
						    const float g,
						    const float b,
						    const float a=1.0f) {
		m_Specular.SetColour(r,g,b,a);
	}

	/// a function to set the emission colour
	inline void SetEmission(const float r,
						    const float g,
						    const float b,
						    const float a=1.0f) {
		m_Emission.SetColour(r,g,b,a);
	}

	/// a function to set the emission colour
	inline void SetTransparency(const float r,
						        const float g,
						        const float b,
						        const float a=1.0f) {
		m_Transparency.SetColour(r,g,b,a);
	}

	/// a function to set the shininess of the material
	inline void SetShininess(const float shine) {
		m_Shininess = shine;
	}

	/// a function to set the texture object for the material to use
	inline void SetTextureStartNum(unsigned int tex) {
		if(m_Ambient.m_TextureObject != 0 ) m_Ambient.m_TextureObject += (tex-1);
		if(m_Diffuse.m_TextureObject != 0 ) m_Diffuse.m_TextureObject += (tex-1);
		if(m_Specular.m_TextureObject != 0 ) m_Specular.m_TextureObject += (tex-1);
		if(m_Emission.m_TextureObject != 0 ) m_Emission.m_TextureObject += (tex-1);
		if(m_Transparency.m_TextureObject != 0 ) m_Transparency.m_TextureObject += (tex-1);
	}

protected:


	class XColorChannel
	{
	public:

		XColorChannel() : m_TextureObject(0) {
			m_Colour[0] = 0;
			m_Colour[1] = 0;
			m_Colour[2] = 0;
			m_Colour[3] = 0;
		}

		inline float Clamp(const float &f) const {
			if (f>1.0f) {
				return 1.0f;
			}	
			if (f<0.0f) {
				return 0.0f;
			}
			return f;
		}
	
		friend std::istream& operator >> (std::istream& ifs,XColorChannel& channel)
		{
			char buffer[256];

			ifs >> buffer // channel name
				>> channel.m_Colour[0] >> channel.m_Colour[1] >> channel.m_Colour[2] >> channel.m_Colour[3]
				>> buffer // Tex=
				>> channel.m_TextureObject;
				return ifs;
		}

		/// texture ID
		unsigned int m_TextureObject;

		/// the RGBA colour
		float m_Colour[4];


		/// a function to set the emission colour
		inline void SetColour(const float r,
							  const float g,
							  const float b,
							  const float a=1.0f) {
			m_Colour[0] = Clamp(r);
			m_Colour[1] = Clamp(g);
			m_Colour[2] = Clamp(b);
			m_Colour[3] = Clamp(a);
		}
	};

	/// the ambient material colour
	XColorChannel m_Ambient;

	/// the diffuse colour
	XColorChannel m_Diffuse;

	/// the specular colour
	XColorChannel m_Specular;

	/// the emission colour
	XColorChannel m_Emission;

	/// the transparency colour
	XColorChannel m_Transparency;

	/// the shininess
	float m_Shininess;

	/// as an optimisation, we will track which material is currently in use. This will 
	/// try to minimise the state changes within OpenGL
	static XMaterial* ms_CurrentMaterial;
};


/// a list of material types
typedef std::vector< XMaterial* > XMaterialList;

/// an iterator type for the material list
typedef XMaterialList::iterator XItMaterial;

//--------------------------------------------------------------------------
/// the phong material type
class XPhong : public XMaterial
{
public:
	friend std::istream& operator >> (std::istream& ifs,XPhong& ph);
};

//--------------------------------------------------------------------------
/// the lambert material type
class XLambert : public XMaterial
{
public:
	friend std::istream& operator >> (std::istream& ifs,XLambert& ph);
};

//--------------------------------------------------------------------------
/// the blinn material type
class XBlinn : public XMaterial
{
public:
	friend std::istream& operator >> (std::istream& ifs,XBlinn& ph);
};

#endif

