//--------------------------------------------------------------------------
/// 
/// \file	Mesh.cpp
/// \author	Rob Bateman
/// \brief	a simple file loader for a simple file... This loader basically 
///			uses streaming operators to read and write the file. The data is 
///			stored in a set os basic classes, XVertex, XNormal, XTexCoord and
///			XFace.
/// 
///			The XMesh class then holds the data within arrays, these array  
///			types are typedefs of std::vector<XVertex> etc.
/// \note       
/// 
/// 		This sourcefile is provided as-is and no guarentee or warrenty 
///			is provided for any damage that may arise from using it. This
///			code is free for your own use, the only thing i ask is a small
///			credit somewhere for my work. An e-mail saying you found it 
///			useful would also be much appreciated by myself ;)
///
///			If you find any bugs arising to maya version changes or you
///			have improvements to the source code, then feel free to mail
///			them to me and i'll update the source-code on the website.
///			Before you e-mail bug reports, make sure that it is my code 
///			causing the problem, and not some wierd oddity in the maya 
///			scene. This code is designed to handle most common model set-ups,
///			it is not designed to replicate perfectly everything that maya
///			does. 
/// 
///			DO NOT send me feature requests, i'm not likely to add to this
///			code myself unless something really new and groovy comes along.
///			This is mainly due to my current time restrictions. If you do
///			have code written that adds something fairly groovy to the 
///			exporter, then i will happily update the exporter code and give
///			full credit where due.....
/// 
///			Robert Bateman
/// 
///				email code to:    rbateman@bmth.ac.uk
///				contact:          robthebloke@hotmail.com
///	
//--------------------------------------------------------------------------

#include "Mesh.h"
#include <assert.h>

//--------------------------------------------------------------------------	XTexCoord

std::istream &operator>>(std::istream& ifs,XTexCoord& v)	{
	return ifs >> v.u >> v.v;
}
bool ReadingUvs=true;

/// streaming operator
std::istream &operator>>(std::istream& ifs,XPointIndex& pi) {
	ifs >> pi.v >> pi.n; 
	if(ReadingUvs) ifs >> pi.t;
	return ifs;
}

//--------------------------------------------------------------------------	XFace
/// \brief	This function sets the number of indices within a face
/// \param	numPoints	-	the number of points within the face
/// 
void XFace::SetSize(unsigned int numPoints) 
{
	// ensure a minimum of 3 points
	if(numPoints<3)
		numPoints = 3;

	// create array of points
	XPointIndex* temp = new XPointIndex[numPoints];

	// copy over old data. Amount to copy depends on whether
	// array has shrunk or increased
	if(m_Points) {
		if (m_NumPoints<numPoints) {
			memcpy(temp,m_Points,sizeof(XVertex)*m_NumPoints);
		} else {
			memcpy(temp,m_Points,sizeof(XVertex)*numPoints);
		}

	}
	m_NumPoints = numPoints;

	// delete old data
	delete [] m_Points;

	// store reference to allocated memory
	m_Points = temp;
}

//--------------------------------------------------------------------------
/// srteaming operator for the face
std::istream &operator>>(std::istream& ifs,XFace& f)	
{
	char buffer[128];
	int vcount;
	ifs >> buffer >> vcount;
	f.SetSize(vcount);

	// read material index for face
	ifs >> buffer >> f.m_Material;

	XPointIndex *p = f.m_Points;
	XPointIndex *end = p + vcount;
	for(;p!=end;++p) 
	{
		ifs >> *p;
	}
	return ifs;
}

//--------------------------------------------------------------------------	XMesh
/// \brief	this function draws all the faces in the mesh
///
void XMesh::gl(XMaterialList& mats) {
	
	XFaceIt it = m_Faces.begin();
	for( ; it != m_Faces.end() ; ++it ) 
	{
		// make sure we have points
		assert(it->m_Points);

		// ensure a valid material ID
		if (it->m_Material != 0 &&  it->m_Material <= mats.size()) {
			// Bind the relevant material
			mats[it->m_Material-1]->gl();
		}

		XPointIndex *p = it->m_Points;
		XPointIndex *end = p + it->m_NumPoints;

		glBegin(GL_POLYGON);
		for(;p!=end;++p) 
		{
			if(m_TexCoords.size())
				m_TexCoords[p->t].gl();
			m_Normals  [p->n].gl();
			m_Vertices [p->v].gl();
		}
		glEnd();
	}
}
//--------------------------------------------------------------------------
/// \brief	stream insertion operator for the mesh class
/// 
std::istream &operator>>(std::istream& ifs,XMesh& m)
{
	char buffer[256];
	int temp;

	// read mesh name
	ifs >> buffer;

	// read vertex count
	ifs >> buffer >> temp;
	m.m_Vertices.resize(temp);

	// read normal count
	ifs >> buffer >> temp;
	m.m_Normals.resize(temp);

	// read tex coord count
	ifs >> buffer >> temp;
	m.m_TexCoords.resize(temp);

	// read face count
	ifs >> buffer >> temp;
	m.m_Faces.resize(temp);

	// read vertices
	ifs >> buffer;
	{
		for( XVertIt it = m.m_Vertices.begin(); it != m.m_Vertices.end() ; ++it ) 
			ifs >> *it;
	}

	// read normals
	ifs >> buffer;
	{
		for( XNormIt it = m.m_Normals.begin(); it != m.m_Normals.end() ; ++it ) 
			ifs >> *it;
	}

	if(m.m_TexCoords.size())
	{
		ReadingUvs=true;
		// read texture coords
		ifs >> buffer;
		for( XTexIt it = m.m_TexCoords.begin(); it != m.m_TexCoords.end() ; ++it ) 
			ifs >> *it;
	}
	else ReadingUvs=false;

	// read the faces
	{
		for( XFaceIt it = m.m_Faces.begin(); it != m.m_Faces.end() ; ++it ) 
			ifs >> *it;
	}

	return ifs;
}



