//--------------------------------------------------------------------------
/// 
/// \file	Model.cpp
/// \author	Rob Bateman
/// \brief	This file defines the XModel class. basicaly this class will load the 
///			contents of a mod file. This version merely holds lights and 
///			meshes in one place and is responsible for drawing them.
/// \note       
/// 
/// 		This sourcefile is provided as-is and no guarentee or warrenty 
///			is provided for any damage that may arise from using it. This
///			code is free for your own use, the only thing i ask is a small
///			credit somewhere for my work. An e-mail saying you found it 
///			useful would also be much appreciated by myself ;)
///
///			If you find any bugs arising to maya version changes or you
///			have improvements to the source code, then feel free to mail
///			them to me and i'll update the source-code on the website.
///			Before you e-mail bug reports, make sure that it is my code 
///			causing the problem, and not some wierd oddity in the maya 
///			scene. This code is designed to handle most common model set-ups,
///			it is not designed to replicate perfectly everything that maya
///			does. 
/// 
///			DO NOT send me feature requests, i'm not likely to add to this
///			code myself unless something really new and groovy comes along.
///			This is mainly due to my current time restrictions. If you do
///			have code written that adds something fairly groovy to the 
///			exporter, then i will happily update the exporter code and give
///			full credit where due.....
/// 
///			Robert Bateman
/// 
///				email code to:    rbateman@bmth.ac.uk
///				contact:          robthebloke@hotmail.com
///	
//--------------------------------------------------------------------------

#include "Model.h"
#include "TextureFile.h"
#include <assert.h>
#include <fstream>

//--------------------------------------------------------------------------
XModel::XModel(void) : m_TextureObject(0),m_NumTextures(0)
{
}

//--------------------------------------------------------------------------
XModel::~XModel(void)
{
	// delete tesxtures
	glDeleteTextures(m_NumTextures,&m_TextureObject);

	// delete all lights
	{
		XItLight it = m_Lights.begin();
		for( ; it != m_Lights.end(); ++it )
		{
			delete *it;
		}
	}

	// delete all meshes
	{
		XItMesh it = m_Meshes.begin();
		for( ; it != m_Meshes.end(); ++it )
		{
			delete *it;
		}
	}
}

//--------------------------------------------------------------------------
/// this Function enables all of the lights in the model
void XModel::EnableLights()
{
	// return false if no lights exist
	if (m_Lights.empty()) {
		return;
	}

	// enable lighting
	glEnable(GL_LIGHTING);

	// if there are no lights, simply enable the default light
	if (m_Lights.size()==0) 
	{
		glEnable(GL_LIGHT0);
	}
	else 
	{
		// enable all lights
		XItLight it = m_Lights.begin();
		for(;it!=m_Lights.end();++it)
		{
			(*it)->Enable();
		}
	}
}

//--------------------------------------------------------------------------
/// this function disables all of the lights in the model
void XModel::DisableLights()
{
	// return if no lights exist
	if (m_Lights.empty()) {
		return;
	}
	
	// disable lighting
	glDisable(GL_LIGHTING);

	// disable each light
	XItLight it = m_Lights.begin();
	for(;it!=m_Lights.end();++it)
	{
		(*it)->Disable();
	}
}

//--------------------------------------------------------------------------
/// a function to draw the geometry data in the model
void XModel::gl()
{
	glPushMatrix();

		EnableLights();

		// draw each mesh
		XItMesh it = m_Meshes.begin();
		for(;it!=m_Meshes.end();++it)
		{
			(*it)->gl(m_Materials);
		}

		DisableLights();

	glPopMatrix();
}

//--------------------------------------------------------------------------
/// \brief	a function to load a file into the model
/// \param	filename	-	the name of the file to load
/// \return	true if OK
/// 
bool XModel::Load(const char* const filename)
{
	{
		// first of all, see if there is a texture file
		std::string texfile = filename;
		texfile.resize(texfile.size()-3);
		texfile += "tex";

		// load the texture file
		XTextureFile texLoader;
		if(texLoader.Load(texfile.c_str()))
		{
			// if OK, pass all textures to openGL
			if(texLoader.TurnIntoTextures(m_TextureObject))
			{
				// store number of textures
				m_NumTextures = texLoader.GetNumTextures();
			}
		} else {
			std::cout << "Could not find " << texfile.c_str() << std::endl;
		}
	}

	char buffer[32];
	std::ifstream ifs(filename);

	if (ifs) {

		while (!ifs.eof()) {
			ifs >> buffer;
			if (ifs.eof()) {
				break;
			}
			std::cout << buffer << std::endl;

			if (buffer[0] == 'B' &&
				buffer[1] == 'L' &&
				buffer[2] == 'I' &&
				buffer[3] == 'N' &&
				buffer[4] == 'N' ) 
			{
				XBlinn* pp = new XBlinn;
				assert (pp);
				ifs >> *pp;
				if (m_TextureObject!=0) {

					pp->SetTextureStartNum( m_TextureObject );
				}
				m_Materials.push_back(pp);
			}
			else
			if (buffer[0] == 'L' )
			{
				if(	buffer[1] == 'I' &&
					buffer[2] == 'G' &&
					buffer[3] == 'H' &&
					buffer[4] == 'T' ) 
				{
					XLight* pl = new XLight( static_cast<unsigned int>(m_Lights.size()));
					assert (pl);
					ifs >> *pl;
					m_Lights.push_back(pl);
				}
				else
				if(	buffer[1] == 'A' &&
					buffer[2] == 'M' &&
					buffer[3] == 'B' &&
					buffer[4] == 'E' &&
					buffer[5] == 'R' &&
					buffer[6] == 'T') 
				{
					XLambert* pp = new XLambert;
					assert (pp);
					ifs >> *pp;
					if (m_TextureObject!=0) {

						pp->SetTextureStartNum( m_TextureObject );
					}
					m_Materials.push_back(pp);
				}
			}
			else
			if (buffer[0] == 'M' &&
				buffer[1] == 'E' &&
				buffer[2] == 'S' &&
				buffer[3] == 'H') 
			{
				XMesh* pm = new XMesh;
				assert (pm);
				ifs >> *pm;
				m_Meshes.push_back(pm);
			}
			else
			if (buffer[0] == 'P' &&
				buffer[1] == 'H' &&
				buffer[2] == 'O' &&
				buffer[3] == 'N' &&
				buffer[4] == 'G' ) 
			{
				XPhong* pp = new XPhong;
				assert (pp);
				ifs >> *pp;
				if (m_TextureObject!=0) {

					pp->SetTextureStartNum( m_TextureObject );
				}
				m_Materials.push_back(pp);
			}
			else
			{
#ifdef DEBUG
				std::cerr << "[ERROR] FileIO Error while trying to read " 
						  << filename << std::endl;
#endif
				ifs.close();
				return false;
			}
		}
		ifs.close();
	}
	return true;
}
