///-----------------------------------------------------------------------------
/// 
/// \file	Texture.h
/// \author	Rob Bateman
/// \brief	This file defines the XImageHeader and the XTexture classes. These 
///			are used to read a texture file from a texture file
/// \note       
/// 
/// 		This sourcefile is provided as-is and no guarentee or warrenty 
///			is provided for any damage that may arise from using it. This
///			code is free for your own use, the only thing i ask is a small
///			credit somewhere for my work. An e-mail saying you found it 
///			useful would also be much appreciated by myself ;)
///
///			If you find any bugs arising to maya version changes or you
///			have improvements to the source code, then feel free to mail
///			them to me and i'll update the source-code on the website.
///			Before you e-mail bug reports, make sure that it is my code 
///			causing the problem, and not some wierd oddity in the maya 
///			scene. This code is designed to handle most common model set-ups,
///			it is not designed to replicate perfectly everything that maya
///			does. 
/// 
///			DO NOT send me feature requests, i'm not likely to add to this
///			code myself unless something really new and groovy comes along.
///			This is mainly due to my current time restrictions. If you do
///			have code written that adds something fairly groovy to the 
///			exporter, then i will happily update the exporter code and give
///			full credit where due.....
/// 
///			Robert Bateman
/// 
///				email code to:    rbateman@bmth.ac.uk
///				contact:          robthebloke@hotmail.com
///	
//--------------------------------------------------------------------------

#ifdef WIN32
#pragma once
#endif

#ifndef __X_TEXTURE__H__
#define __X_TEXTURE__H__

#include <stdio.h>
#include <math.h>

//--------------------------------------------------------------------------
/// \brief	This struct is used as a 2byte identifier just before an image
/// \note	sizeof(ImageHeader) == 2
/// 
struct XImageHeader
{
	/// the width of the image as the power of two
	unsigned char m_WidthPower:4;
	/// the height of the image as a power of two
	unsigned char m_HeightPower:4;

	/// the bits per pixel
	/// 0 = 8bit
	/// 1 = 16bit
	/// 2 = 24bit
	/// 3 = 32bit
	/// 
	unsigned char m_Bpp:2;

	/// 6bits for a rainy day...
	unsigned char m_Unused:6;
};

//--------------------------------------------------------------------------
/// \brief	a class to hold a texture
/// 
class XTexture
{
public:
	
	/// ctor
	XTexture(void);

	/// dtor
	~XTexture(void);

	/// returns the width of the texture
	inline unsigned short GetWidth() const {
		return static_cast<unsigned short>(pow(static_cast<unsigned short>(2),m_Header.m_WidthPower));
	}

	/// returns the height of the texture
	inline unsigned short GetHeight() const {
		return static_cast<unsigned short>(pow(static_cast<unsigned short>(2),m_Header.m_HeightPower));
	}

	/// returns the bits per pixel of the texture
	inline unsigned short GetBitsPerPixel() const {
		return GetBytesPerPixel() << 3;
	}

	/// returns the bytes per pixel of the texture
	inline unsigned short GetBytesPerPixel() const {
		return 1 + m_Header.m_Bpp;
	}

	/// returns the bytes per pixel of the texture
	inline const unsigned char* const GetPixels() const {
		return m_Pixels;
	}

	/// this function reads a texture from the specified file pointer
	bool Read(FILE* fp);

protected:

	/// the image header
	XImageHeader   m_Header;

	/// a pointer to the pixel data
	unsigned char* m_Pixels;
};

#endif


