///-----------------------------------------------------------------------------
/// 
/// \file	TextureFile.cpp
/// \author	Rob Bateman
/// \brief	the XTextureFile class holds a list of all loaded textures from the 
///			specified file.
/// \note       
/// 
/// 		This sourcefile is provided as-is and no guarentee or warrenty 
///			is provided for any damage that may arise from using it. This
///			code is free for your own use, the only thing i ask is a small
///			credit somewhere for my work. An e-mail saying you found it 
///			useful would also be much appreciated by myself ;)
///
///			If you find any bugs arising to maya version changes or you
///			have improvements to the source code, then feel free to mail
///			them to me and i'll update the source-code on the website.
///			Before you e-mail bug reports, make sure that it is my code 
///			causing the problem, and not some wierd oddity in the maya 
///			scene. This code is designed to handle most common model set-ups,
///			it is not designed to replicate perfectly everything that maya
///			does. 
/// 
///			DO NOT send me feature requests, i'm not likely to add to this
///			code myself unless something really new and groovy comes along.
///			This is mainly due to my current time restrictions. If you do
///			have code written that adds something fairly groovy to the 
///			exporter, then i will happily update the exporter code and give
///			full credit where due.....
/// 
///			Robert Bateman
/// 
///				email code to:    rbateman@bmth.ac.uk
///				contact:          robthebloke@hotmail.com
///	
//--------------------------------------------------------------------------

#include "TextureFile.h"

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>
#endif
#include <GL/gl.h>
#include <iostream>

//--------------------------------------------------------------------------
XTextureFile::XTextureFile(void)
{
}

//--------------------------------------------------------------------------
XTextureFile::~XTextureFile(void)
{
	for(iterator it=begin(); it != end(); ++it)
	{
		delete *it;
	}
}

//--------------------------------------------------------------------------
void MakeGlTexture(unsigned int id,GLenum Format,const unsigned char *pData,unsigned int iW,unsigned int iH)
{
	glPixelStorei (GL_UNPACK_ALIGNMENT, 0);
	glBindTexture (GL_TEXTURE_2D, id);

	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D (GL_TEXTURE_2D,0,Format,iW,iH,0,Format,GL_UNSIGNED_BYTE,pData);
}

//--------------------------------------------------------------------------
/// this function converts all image data into a single openGL texture object. 
/// to use each texture, use tex_obj + n to get the n'thy texture object
/// 
bool XTextureFile::TurnIntoTextures(unsigned int& tex_obj)
{
	if (GetNumTextures() != 0) {
		glGenTextures(GetNumTextures(),&tex_obj);

		unsigned int temp=tex_obj;
		// for each texture
		for(iterator it=begin(); it != end(); ++it,++temp)
		{
			GLenum Format;
			switch((*it)->GetBytesPerPixel()) {

			// if 8bit greyscale, make alpha texture
			case 1:
				Format = GL_ALPHA;
				break;

			// if 16bit
			case 2:
				Format = GL_RGB5_A1;
				break;

			// if 24 bit
			case 3:
				Format = GL_RGB;
				break;

			// if 32 bit
			case 4:
				Format = GL_RGBA;
				break;

			// err? shouldn't ever get here
			default:
#ifdef DEBUG
				std::cerr << "Unable to handle the specified bytes size" << std::endl;
#endif
				break;
			}
            MakeGlTexture(temp,Format,(*it)->GetPixels(),(*it)->GetWidth(),(*it)->GetHeight());			
		}
		return true;
	}
	return false;
}

//--------------------------------------------------------------------------
/// this function loads all of the specified textures from the specified file
/// 
bool XTextureFile::Load(const char* const filename)
{
	// open the file
	FILE *fp = fopen(filename,"rb");
	if (fp) {

		// read number of textures contained
        unsigned short numTextures;
		fread(&numTextures,2,1,fp);

		// resize the array, set all pointers to NULL
		resize(numTextures,NULL);

		// for each texture
		for(iterator it=begin(); it != end(); ++it)
		{
			// allocate new texture
			*it = new XTexture;

			// read the image data
			if(!(*it)->Read(fp))
			{
				// if failed delete node, and return
				delete *it;
				*it = NULL;
				fclose(fp);
				return false;
			}
		}
		fclose(fp);
		return true;
	}
	return false;
}

