// EulerAngles.cpp: implementation of the XEulerAngles class.
//
//////////////////////////////////////////////////////////////////////

#include "EulerAngles.h"
#include <cassert>
#ifndef USE_ROBS_OPTIMISED_STUPH
#define USE_ROBS_OPTIMISED_STUPH 0
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

XMatrix XEulerAngles::AsMatrix() const
{
// building up euler angled matrices in a certain rotation order can be
// a slow process. To speed things up a little bit, you can pre-calculate 
// the matrix multipications required and ignore those calculations that return
// zero or one. If using matrices to build the final matrix, you would require
// 2 sin & 2 cos calcs per rotation axis, thus 6
// 2 4x4 matrix calculations, a total of 128 multiplications, 96 additions
//
// if we work out the matrix math beforehand and ignore all values we know to
// return 0 or 1, we can condense this to 
//
// 2 sin & 2 cos calcs per rotation axis, thus 6
// 
//   12 multiplications,  4 additions
//
// That's less than 10% of the matrix method !
// 
// I have also included the matrix method purely for verification of my rusty maths skills
//
#if USE_ROBS_OPTIMISED_STUPH
	/*
		Given the rotation matrices:

			 |  1  0  0 |
		X  = |  0  A  B |
			 |  0 -B  A |

			 |  C  0  D |
		Y  = |  0  1  0 |
			 | -D  0  C |

			 |  E  F  0 |
		Z  = | -F  E  0 |
			 |  0  0  1 |

		where A,B are the cosine and sine of the X-axis rotation,
			  C,D are the cosine and sine of the Y-axis rotation,
			  E,F are the cosine and sine of the Z-axis rotation.

	*/

	// pre-calculate sin and cosine values

	XReal A = cosf(m_xRotation);
	XReal B = sinf(m_xRotation);

	XReal C = cosf(m_yRotation);
	XReal D = sinf(m_yRotation);

	XReal E = cosf(m_zRotation);
	XReal F = sinf(m_zRotation);

	// depending on the order in which they are multiplied....
	switch(m_RotationOrder)
	{
	case XROT_XYZ:
		{
	  
			/*
			
				 X.Y.Z	=  |  1  0  0 |     |  C  0  D |     |  E  F  0 |
						   |  0  A  B |  x  |  0  1  0 |  x  | -F  E  0 |
						   |  0 -B  A |     | -D  0  C |     |  0  0  1 |

						=  |   C  0   D |     |  E  F  0  |
						   | -BD  A  BC |  X  | -F  E  0  |
						   |  AD -B  AC |     |  0  0  1  |

						=  |      CE      CF   C  |
						   | -BDE-AF -BDF+AE  BC  |
						   |  ADE+BF  ADF-BE  AC  |
			*/

			XReal AC = C * A;
			XReal BC = C * B;
			XReal CE = C * E;
			XReal CF = C * F;

			XReal AE = A * E;
			XReal AF = A * F;
			XReal BE = B * E;
			XReal BF = B * F;

			XReal ADE = AE * D;
			XReal ADF = AF * D;
			XReal BDE = BE * D;
			XReal BDF = BF * D;

			return XMatrix(       CE,      CF,  C,  0, 
							 -BDE-AF, -BDF+AE, BC,  0,
							  ADE+BF,  ADF-BE, AC,  0,
								   0,       0,  0,  1 );
		}
	case XROT_XZY:
		{
			/*
				X.Z.Y	=  |  1  0  0 |     |  E  F  0 |     |  C  0  D |
						   |  0  A  B |  x  | -F  E  0 |  x  |  0  1  0 |
						   |  0 -B  A |     |  0  0  1 |     | -D  0  C |

						=  |   E   F  0  |     |  C  0  D  |
						   | -AF  AE  B  |  X  |  0  1  0  |
						   |  BF -BE  A  |     | -D  0  C  |

						=  |      CE    F       DE |
						   | -ACF-BD   AE  -ADF+BC |
						   |  BCF-AD  -BE   BDF+AC |

			*/

			XReal AE = A * E;
			XReal BE = B * E;
			XReal CE = C * E;
			XReal DE = D * E;

			XReal AC = A * C;
			XReal AD = A * D;
			XReal BC = B * C;
			XReal BD = B * D;

			XReal ACF = AC * F;
			XReal ADF = AD * F;
			XReal BCF = BC * F;
			XReal BDF = BD * F;

			return XMatrix(       CE,       F,      DE,  0, 
							 -ACF-BD,      AE, -ADF+BC,  0,
							  BCF-AD,     -BE,  BDF+AC,  0,
								   0,       0,       0,  1 );
		}
	

	case XROT_YXZ:
		{
			/*
				Y.X.Z	=  |  C  0  D |     |  1  0  0 |     |  E  F  0 |
						   |  0  1  0 |  x  |  0  A  B |  x  | -F  E  0 |
						   | -D  0  C |     |  0 -B  A |     |  0  0  1 |

						=  |  C  0  D  |     |   E   F  0  |     
						   |  0  1  0  |  X  | -AF  AE  B  |
						   | -D  0  C  |     |  BF -BE  A  |     

						=  |  CE+BDF  CF-BDE    AD   |
						   |  -AF         AE     B   |
						   | -DE+BCF -DF-BCE    AC   |

			*/
			XReal AC = A * C;
			XReal AD = A * D;
			XReal AE = A * E;
			XReal AF = A * F;

			XReal CE = C * E;
			XReal CF = C * F;
			XReal DE = D * E;
			XReal DF = D * F;

			XReal BCE = CE * B;
			XReal BCF = CF * B;
			XReal BDE = DE * B;
			XReal BDF = DF * B;

			return XMatrix(   CE+BDF,  CF-BDE,      AD,  0, 
							     -AF,      AE,       B,  0,
							 -DE+BCF, -DF-BCE,      AC,  0,
								   0,       0,       0,  1 );
		}

	


	case XROT_YZX:
		{
			/*
				Y.Z.X	=  |  C  0  D |     |  E  F  0 |     |  1  0  0 |  
						   |  0  1  0 |  x  | -F  E  0 |  x  |  0  A  B |
						   | -D  0  C |     |  0  0  1 |     |  0 -B  A |     

						=  |  C  0  D  |     |   E  AF  BF |     
						   |  0  1  0  |  x  |  -F  AE  BE |
						   | -D  0  C  |     |   0  -B   A |     

						=  |   CE   ACF-BD   BCF+AD  |
						   |   -F       AE       BE  |
						   |                         |

			*/

			XReal AE = A * E;
			XReal BE = B * E;
			XReal CE = C * E;
			XReal DE = D * E;

			XReal AC = A * C;
			XReal AD = A * D;
			XReal BC = B * C;
			XReal BD = B * D;

			XReal ACF = AC * F;
			XReal ADF = AD * F;
			XReal BCF = BC * F;
			XReal BDF = BD * F;

			return XMatrix(       CE,       F,      DE,  0, 
							 -ACF-BD,      AE, -ADF+BC,  0,
							  BCF-AD,     -BE,  BDF+AC,  0,
								   0,       0,       0,  1 );
		}
	


	case XROT_ZXY:
		{
			/*
			
				  Z.X.Y	=  |  E  F  0 |     |  1  0  0 |     |  C  0  D |
						   | -F  E  0 |  x  |  0  A  B |  x  |  0  1  0 |
						   |  0  0  1 |     |  0 -B  A |     | -D  0  C |     

						=  |  E  F  0  |     |   C  0   D |     
						   | -F  E  0  |  X  | -BD  A  BC | 
						   |  0  0  1  |     |  AD -B  AC |     

						=  |  CE-BDF   AF    DE+BCF  |
						   | -CF-BDE   AE   -DF+BCE  |
						   |      AD   -B        AC  |
			*/


			XReal AC = A * C;
			XReal AD = A * D;
			XReal AE = A * E;
			XReal AF = A * F;

			XReal CE = B * E;
			XReal CF = B * F;
			XReal DE = D * E;
			XReal DF = D * F;

			XReal BCE = CE * B;
			XReal BCF = CF * B;
			XReal BDE = DE * B;
			XReal BDF = DF * B;

			return XMatrix(   CE-BDF,      AF,  DE+BCF,  0, 
							 -CF-BDE,      AE, -DF+BCE,  0,
								  AD,      -B,      AC,  0,
								   0,       0,       0,  1 );
		}


	case XROT_ZYX:
		{
			/*
			
				  Z.Y.X	=  |  E  F  0 |     |  C  0  D |     |  1  0  0 |
						   | -F  E  0 |  x  |  0  1  0 |  x  |  0  A  B |
						   |  0  0  1 |     | -D  0  C |     |  0 -B  A |     

						=  |  E  F  0  |     |   C -BD  AD |     
						   | -F  E  0  |  X  |   0   A   B | 
						   |  0  0  1  |     |  -D -BC  AC |     

						=  |  CE    -BDE+AF   ADE+BF |
						   | -CF     BDF+AE  -ADF+BE |
						   |  -D        -BC       AC |
			*/

			XReal AC = C * A;
			XReal BC = C * B;
			XReal CE = C * E;
			XReal CF = C * F;

			XReal AE = A * E;
			XReal AF = A * F;
			XReal BE = B * E;
			XReal BF = B * F;

			XReal ADE = AE * D;
			XReal ADF = AF * D;
			XReal BDE = BE * D;
			XReal BDF = BF * D;

			return XMatrix(   CE,    -BDE+AF,  ADE+BF,  0, 
							 -CF,     BDF+AE, -ADF+BE,  0,
							  -D,        -BC,      AC,  0,
							   0,          0,       0,  1 );
		}
	default:break;
	}

#else

	// this bit of code performs the calculation for the rotation using the
	// very un-optimal method of multiplying matrices. This is purely used for
	// my development purposes so I can test the above optimised routines. I left
	// it in incase anyone finds fault with my calculations or wants to verify my
	// code for me.....
	//
	XMatrix rotx,roty,rotz;

	// perform rotations on individual matrices
	rotx.RotateX(m_xRotation);
	roty.RotateY(m_yRotation);
	rotz.RotateZ(m_zRotation);

	// build final rotation dependant on rotation order
	switch(m_RotationOrder)
	{
	case XROT_XYZ:
		{
			XMatrix m;
			m *= rotx;
			m *= roty;
			m *= rotz;
			return m;
		}

	case XROT_XZY:
		{
			XMatrix m;
			m *= rotx;
			m *= rotz;
			m *= roty;
			return m;
		}
	
	case XROT_YZX:
		{
			XMatrix m;
			m *= roty;
			m *= rotz;
			m *= rotx;
			return m;
		}
	
	case XROT_YXZ:
		{
			XMatrix m;
			m *= roty;
			m *= rotx;
			m *= rotz;
			return m;
		}
	
	case XROT_ZXY:
		{
			XMatrix m;
			m *= rotz;
			m *= rotx;
			m *= roty;
			return m;
		}
	
	case XROT_ZYX:
		{
			XMatrix m;
			m *= rotz;
			m *= roty;
			m *= rotx;
			return m;
		}
	
	default:
		assert(0 && "[XEulerAngles::AsMatrix] Invalid rotation order!");
		break;
	}
	return XMatrix();

#endif

}
