#ifndef _INTERPOLATION__H__
#define _INTERPOLATION__H__

#ifdef WIN32
#pragma once
#endif

#include "MathDefs.h"


///---------------------------------------------------------------------------------
/// \brief	The XLerp template is an experimental class to interpolate between two
///			variables of the same type, V0 and V1, with respect to the blending factor t.
///			The equation for the interpolation is 
/// 
///					V(t)  =   V0 * (1-t) + V1 * t
/// \arg	T	-	the type of variable to lerp
/// 
template
<
	typename T
>
class XLerp
{	
public:

	/// \brief	ctor
	XLerp()
		: V0() , 
		  V1(),
		  t(0)
	{
	}

	/// \brief	ctor
	/// \param	v0	-	the first value
	/// \param	v1	-	the second value
	/// \param	b	-	the blending value
	/// 
	XLerp(const T& v0,
		  const T& v1,
		  const XReal b =0.0f) 
		: V0(v0),
		  V1(v1),
		  t(b)
	{
	}

	/// \brief	The copy constructor
	/// \param	rhs	-	the right hand side to copy
	/// 
	XLerp(const XLerp<T>& rhs) 
		: V0(rhs.V0),
		  V1(rhs.V1),
		  t(rhs.t) 
	{
	}

	/// \brief	This function evaluates and returns the answer to the interpolation
	/// \return	The result of the interpolation
	/// 
	inline const T Evaluate() const {
		return (V0*(1.0f-t)) + V1*t;
	}

	/// \brief	This operator is called when the object is read like a variable of type T.
	///			It essentually returns the result of the calculation
	/// \return	The evaluated answer
	/// 
	const T operator ()() const {
		return Evaluate();
	}

	/// the first interpolation value
	T		V0;

	/// the second interpolation value
	T		V1;

	/// the blending factor
	XReal	t;
};


///---------------------------------------------------------------------------------
/// \brief	
/// \arg	
/// 
template< typename T >
class XQuadraticLerp
{	
public:

	XQuadraticLerp()
		: V0(), 
		  V1(),
		  V2(),
		  t(0)
	{
	}
	XQuadraticLerp(const T& v0,const T& v1,const T& v2,const XReal b =0.0f) 
		: V0(v0),
		  V1(v1),
		  V2(v2),
		  t(b)
	{
	}
	XQuadraticLerp(const XQuadraticLerp<T>& rhs) 
		: V0(rhs.V0),
		  V1(rhs.V1),
		  V2(rhs.V2),
		  t(rhs.t) 
	{
	}

	T operator ()() {
		XReal invt = (1.0f-t);
		XReal invt2 = invt*invt;
		XReal t2 = t*t;
		return (V0*invt2) + (V1*2*t*invt) + (V2*t2);
	}

	/// the first interpolation value
	T		V0;

	/// the second interpolation value
	T		V1;

	/// the third interpolation value
	T		V2;

	/// the blending factor
	XReal	t;
};

///---------------------------------------------------------------------------------
/// \brief	
/// \arg	
/// 
template< typename T >
class XBezierLerp
{	
public:

	XBezierLerp()
		: V0(), 
		  V1(),
		  V2(),
		  V3(),
		  t(0)
	{
	}
	XBezierLerp(const T& v0,const T& v1,const T& v2,const T& v3,const XReal b =0.0f) 
		: V0(v0),
		  V1(v1),
		  V2(v2),
		  V3(v3),
		  t(b)
	{
	}
	XBezierLerp(const XBezierLerp<T>& rhs) 
		: V0(rhs.V0),
		  V1(rhs.V1),
		  V2(rhs.V2),
		  V3(rhs.V3),
		  t(rhs.t) 
	{
	}

	T operator ()() {
		XReal invt = 1.0f-t;
		XReal invt2 = invt*invt;
		XReal invt3 = invt2*invt;
		XReal t2 = t*t;
		XReal t3 = t2*t;
		return V0*invt3 + V1*3*t*invt2 + V2*3*t2*invt2 + V3*t3;
	}

	/// the first interpolation value
	T		V0;

	/// the second interpolation value
	T		V1;

	/// the third interpolation value
	T		V2;

	/// the fourth interpolation value
	T		V3;

	/// the blending factor
	XReal	t;
};

///---------------------------------------------------------------------------------
/// \brief	
/// \arg	
/// \param	
/// \param	
/// \param	
/// \param	
/// 
template< typename T >
void LERP(OUT T& result,IN const float t,IN const T& v0,IN const T& v1) {
	result = v0*(1.0f-t) + v1*t;
}
///---------------------------------------------------------------------------------
/// \brief	
/// \arg	
/// \param	
/// \param	
/// \param	
/// 
template< typename T >
T LERP(IN const float t,IN const T& v0,IN const T& v1) {
	return v0*(1.0f-t) + v1*t;
}


#endif




