//=========================================================================
///
///	\file	Matrix.h
/// \author	Rob
///	\brief	Defines the class XMatrix, a class definition for a 4x4 matrix
///
//=========================================================================


//=========================================================================
//	Sentry
//=========================================================================

//	This directive is used to stop the header file from being included
//	Twice and is standard for all headers.
#ifdef WIN32
#pragma once
#endif


#ifndef MATRIX_H_INCLUDED
#define MATRIX_H_INCLUDED

//=========================================================================
//	Includes
//=========================================================================

#include <iostream>
#include <stdlib.h>
#include "MathDefs.h"
#include "Vector.h"
#include "Quaternion.h"

//=========================================================================
//	Class Definitions
//=========================================================================

class XMatrix
{
public:

	//=====================================================================	Data


	//	use a union to offer two ways of accessing
	//	the same chunk of memory.
	//
	union
	{
		/// the member data as an array
		XReal data[16];

		/// using a struct to package together the member data as a
		/// set of individual components.
		struct
		{
			XReal m_00; XReal m_01; XReal m_02; XReal m_03;
			XReal m_10; XReal m_11; XReal m_12; XReal m_13;
			XReal m_20; XReal m_21; XReal m_22; XReal m_23;
			XReal m_30; XReal m_31; XReal m_32; XReal m_33;
		};
	};

	//=====================================================================	Constructors / Destructor

	/// \brief	constructor
	///
	XMatrix();

	/// \brief	copy constructor
	/// \param	m	-	the matrix to copy
	///
	XMatrix(IN const XMatrix &m);

	/// \brief	constructor. 
	/// \note	I really don't like writing functions with this many arguments,
	///			it tends to prevent you using fast call, it is however mildly useful
	///			in some occasions.
	///
	XMatrix(IN XReal _00, IN XReal _01, IN XReal _02, IN XReal _03,
			IN XReal _10, IN XReal _11, IN XReal _12, IN XReal _13,
			IN XReal _20, IN XReal _21, IN XReal _22, IN XReal _23,
			IN XReal _30, IN XReal _31, IN XReal _32, IN XReal _33)
	{
		 m_00 = _00;  m_01 = _01;  m_02 = _02;  m_03 = _03;
		 m_10 = _10;  m_11 = _11;  m_12 = _12;  m_13 = _13;
		 m_20 = _20;  m_21 = _21;  m_22 = _22;  m_23 = _23;
		 m_30 = _30;  m_31 = _31;  m_32 = _32;  m_33 = _33;
	}

	/// \brief	dtor
	~XMatrix();
	
	inline void Identity() {
		memset(data,0,16*sizeof(float));
		m_00 = m_11 = m_22 = m_33 = 1.0f;
	}



	//=====================================================================	Functions

	/// \brief	This function transposes the matrix 
	/// \return	The trasposed version of this matrix
	inline void Transpose() {
		XMatrix mm( m_00, m_10, m_20, m_30, 
		  			m_01, m_11, m_21, m_31,
					m_02, m_12, m_22, m_32,
					m_03, m_13, m_23, m_33 );
					
		memcpy(data,mm.data,16*sizeof(float));
	}
	inline XMatrix Transposed() const {
		return XMatrix( m_00, m_10, m_20, m_30, 
						m_01, m_11, m_21, m_31,
						m_02, m_12, m_22, m_32,
						m_03, m_13, m_23, m_33 );
	}


	///	\brief	a function to action a rotation of the specified
	///			degrees around the x-axis
	/// \param	degrees	-	the amount of the rotation
	/// \return	reference to this for concatonation of function calls
	/// \note
	///
	XMatrix& RotateX( IN const XReal degrees );

	///	\brief	a function to action a rotation of the specified
	///			degrees around the y-axis
	/// \param	degrees	-	the amount of the rotation
	/// \return	reference to this for concatonation of function calls
	/// \note
	///
	XMatrix& RotateY( IN const XReal degrees );

	///	\brief	a function to action a rotation of the specified
	///			degrees around the z-axis
	/// \param	degrees	-	the amount of the rotation
	/// \return	reference to this for concatonation of function calls
	/// \note
	///
	XMatrix& RotateZ( IN const XReal degrees );

	///	\brief	a function to action a scale of (x_,y_,z_)
	/// \param	x_	-	the x-component of the scale
	/// \param	y_	-	the y-component of the scale
	/// \param	z_	-	the z-component of the scale
	/// \return	reference to this for concatonation of function calls
	/// \note
	///
	XMatrix& Scale( IN const XReal x_ ,
				    IN const XReal y_ ,
					IN const XReal z_ );

	///	\brief	a function to action a scale of ( sv.x_, sv.y_, sv.z_ )
	/// \param	sv	-	the scale as a vector
	/// \return	reference to this for concatonation of function calls
	/// \note
	///
	XMatrix& Scale( IN const XVector &sv);

	/// \brief	This function retrieves the Scale given by this matrix
	/// \param	x	-	the output x translation value
	/// \param	y	-	the output y translation value
	/// \param	z	-	the output z translation value
	/// 
	void GetScale(OUT XReal& x_,OUT XReal& y_, OUT XReal& z_) const {
		x_ = m_00;
		y_ = m_11;
		z_ = m_22;
	}

	/// \brief	This function retrieves the Scale given by this matrix
	/// \param	v	-	the output translation value
	/// 
	void GetScale(OUT XVector& v) const {
		v.x = m_00;
		v.y = m_11;
		v.z = m_22;
	}

	///	\brief	a function to action a translation of ( x_, y_, z _)
	/// \param	x_	-	the x-component of the translation
	/// \param	y_	-	the y-component of the translation
	/// \param	z_	-	the z-component of the translation
	/// \return	reference to this for concatonation of function calls
	/// \note
	///
	XMatrix& Translate( IN const XReal x_ ,
						IN const XReal y_ ,
						IN const XReal z_ );

	///	\brief	a function to action a translation of ( tv.x_, tv.y_, tv.z_ )
	/// \param	tv	-	the translation as a vector
	/// \return	reference to this for concatonation of function calls
	/// \note
	///
	XMatrix& Translate( IN const XVector &tv );

	/// \brief	This function retrieves the translation given by this matrix
	/// \param	x	-	the output x translation value
	/// \param	y	-	the output y translation value
	/// \param	z	-	the output z translation value
	/// 
	inline void GetTranslation(OUT XReal& x_,OUT XReal& y_, OUT XReal& z_) const {
		x_ = m_03;
		y_ = m_13;
		z_ = m_23;
	}

	/// \brief	This function retrieves the translation given by this matrix
	/// \param	v	-	the output translation value
	///
	inline void GetTranslation(OUT XVector& v) const {
		v.x = m_13;
		v.y = m_23;
		v.z = m_33;
	}

	/// \brief	This function returns the inverse of this matrix. This only works
	///			if your matrix is affine, ie has rotations and translations only. 
	///			for most animation purposes this should be fairly happy, for more
	///			hectic stuff it will not work and use the full matrix inverse 
	///			function instead
	/// \param	result	-	this will hold the matrix inverse
	///	\return	The inverse of this matrix
	/// 
	void CheapInverse(OUT XMatrix& result) ;

	/// \brief	This function returns the inverse of this matrix. 
	/// \param	result	-	this will hold the matrix inverse
	///	\return	true if inverse could be found
	/// 
	bool Inverse(OUT XMatrix& result) const;
	
	/// 
	bool Inverse() {
		XMatrix inv;
		bool b = Inverse(inv);
		*this = inv;
		return b;
	}

	XQuaternion ToQuaternion() const;

	/// \brief	This function returns determinant of the matrix
	/// \return	The determinant
	/// 
	XReal Determinant() const;
	
	/// \brief	This function pulls out a sub-matrix from the 4x4 matrix
	/// \param	i	-	the x-col to ignore
	/// \param	j	-	the j column to ignore
	/// \param	mat	-	The output matrix
	/// 
	void SubMatrix3x3(IN int i,IN int j,OUT XReal mat[]) const;


	//=====================================================================	Operators

	///	\brief	operator  std::istream >> XMatrix
	///	\param	ifs	-	the input file stream, ie std::cin
	///	\param	m	-	the matrix instance into which the data is being placed.
	/// \return	the stream reference
	/// \note	The stream extraction operator has to be declared as a friend of the
	///			class. This is because the operator itself cannot be a member function
	///			of XMatrix because the type, std::istream, appears on the left of the
	///			operator. For it to be a member function, the XMatrix would have to be
	///			on the left of the >> operator. This forces us to make the operator act
	///			globally and should therefore always be declared as a friend.
	///
	friend std::istream &operator >> (IN std::istream &ifs,IN XMatrix &m);

	///	\brief	operator  std::ostream << XMatrix
	///	\param	ofs	-	the output file stream, ie std::cout, std::fstream etc
	///	\param	m	-	the data instance to output
	/// \return	the stream reference
	/// \note	The stream insertion operator has to be declared as a friend of the
	///			class. This is because the operator itself cannot be a member function
	///			of XMatrix because the type, std::ostream, appears on the left of the
	///			operator. For it to be a member function, the XMatrix would have to be
	///			on the left of the << operator. This forces us to make the operator act
	///			globally and should therefore always be declared as a friend.
	///
	friend std::ostream &operator << (IN std::ostream &ofs,IN XMatrix &m);

	///	\brief	operator == XMatrix
	///	\param	m	-	compare m against *this
	/// \return	true if (more or less within accuracy of g_vecEpsilon)
	///			the same as this, false otherwise
	///
	inline const bool operator==(IN const XMatrix &m) const
	{
		// guess all matrix elements are the same
		return 	FCOMPARE(m_00  , m.m_00) &&
				FCOMPARE(m_01  , m.m_01) &&
				FCOMPARE(m_02  , m.m_02) &&
				FCOMPARE(m_03  , m.m_03) &&

				FCOMPARE(m_10  , m.m_10) &&
				FCOMPARE(m_11  , m.m_11) &&
				FCOMPARE(m_12  , m.m_12) &&
				FCOMPARE(m_13  , m.m_13) &&

				FCOMPARE(m_20  , m.m_20) &&
				FCOMPARE(m_21  , m.m_21) &&
				FCOMPARE(m_22  , m.m_22) &&
				FCOMPARE(m_23  , m.m_23) &&

				FCOMPARE(m_30  , m.m_30) &&
				FCOMPARE(m_31  , m.m_31) &&
				FCOMPARE(m_32  , m.m_32) &&
				FCOMPARE(m_33  , m.m_33) ;
	}

	///	\brief	operator != XMatrix
	///	\param	m	-	compare m against *this
	/// \return	false if (more or less within accuracy of g_vecEpsilon)
	///			the same as this
	///
	inline const bool operator!=(IN const XMatrix &m) const
	{
		return !(m == *this);
	}

	///	\brief	operator =	XMatrix
	///	\param	rhs	-	the matrix on the right hand side of the operator
	/// \return	a constant reference to this
	///
	inline const XMatrix& operator=(IN const XMatrix &rhs)
	{
		m_00  = rhs.m_00;
		m_01  = rhs.m_01;
		m_02  = rhs.m_02;
		m_03  = rhs.m_03;

		m_10  = rhs.m_10;
		m_11  = rhs.m_11;
		m_12  = rhs.m_12;
		m_13  = rhs.m_13;

		m_20  = rhs.m_20;
		m_21  = rhs.m_21;
		m_22  = rhs.m_22;
		m_23  = rhs.m_23;

		m_30  = rhs.m_30;
		m_31  = rhs.m_31;
		m_32  = rhs.m_32;
		m_33  = rhs.m_33;

		return *this;
	}

	///	\brief	operator *=	float
	///	\param	f	-	the float to multiple this by
	/// \return	a constant reference to this
	///
	inline const XMatrix& operator*=(IN const float &f)
	{
		XReal *pEnd = data + 16;
		for( XReal *p = data; p != pEnd; ++p ) {
			( *p ) *= static_cast<XReal>( f );
		}
		return *this;
	}

	///	\brief	operator /=	float
	///	\param	f	-	the float to divide this by
	/// \return	a constant reference to this
	///
	inline const XMatrix& operator /= (IN const float &f)
	{
		XReal *pEnd = data + 16;
		for( XReal *p = data; p != pEnd; ++p ) {
			( *p ) /= static_cast<XReal>( f );
		}
		return *this;
	}

	///	\brief	operator *=	double
	///	\param	d	-	the double to multiple this by
	/// \return	a constant reference to this
	///
	inline const XMatrix& operator*=(IN const double &d)
	{
		XReal *pEnd = data + 16;
		for( XReal *p = data; p != pEnd; ++p ) {
			( *p ) *= static_cast<XReal>( d );
		}
		return *this;
	}

	///	\brief	operator /=	double
	///	\param	d	-	the double to divide this by
	/// \return	a constant reference to this
	///
	inline const XMatrix& operator /= (IN const double &d)
	{
		XReal *pEnd = data + 16;
		for( XReal *p = data; p != pEnd; ++p ) {
			( *p ) /= static_cast<XReal>( d );
		}
		return *this;
	}

	///	\brief	operator *=	XMatrix
	///	\param	m	-	the matrix to multiple this by
	/// \return	a constant reference to this
	///
	const XMatrix& operator *= (IN const XMatrix &m) ;

	///	\brief	operator	XMatrix  *	XMatrix
	///	\param	m	-	the matrix to multiple this by
	/// \return	a constant reference to this
	///
	XMatrix  operator * (IN const XMatrix &m) const;

	///	\brief	operator	XMatrix  *	XVector
	///	\param	v	-	the vector to transform
	/// \return	a constant reference to this
	///
	XVector  operator * (IN const XVector &v) const;

	///	\brief	operator	XVector	*	XMatrix
	///	\param	v	-	the vector to transform
	///	\param	m	-	the matrix to multiple this by
	/// \return	a constant reference to this
	///
	friend XVector  operator * (IN const XVector &v,IN const XMatrix &m);


};


#endif
