//
//	===========================================================================
//	=================  ===========  ============  =====  =========  ===========
//	=================  ===========  ============  =====  =========  ===========
//	=================  ======  ===  ============  =====  =========  ===========
//	=  =   ====   ===  =====    ==  ======   ===  =====  ===   ===  =  ===   ==
//	=    =  ==     ==    ====  ===    ===  =  ==    ===  ==     ==    ===  =  =
//	=  =======  =  ==  =  ===  ===  =  ==     ==  =  ==  ==  =  ==   ====     =
//	=  =======  =  ==  =  ===  ===  =  ==  =====  =  ==  ==  =  ==    ===  ====
//	=  =======  =  ==  =  ===  ===  =  ==  =  ==  =  ==  ==  =  ==  =  ==  =  =
//	=  ========   ===    ====   ==  =  ===   ===    ===  ===   ===  =  ===   ==
//	===========================================================================
//

#include <math.h>
#include "Quaternion.h"

OUT XMatrix XQuaternion::AsMatrix() const {
	XMatrix m;
	AsMatrix(m.data);
	return m;
}


//--------------------------------------------------------------------------
/// \brief	This function converts an axis angle rotation to a quaternion
/// \param	Axis	-	the axis around which the rotation occurs
/// \param	Angle	-	the amount of rotation
///
void XQuaternion::FromAxisAngle(const XVector& Axis,const float Angle) {
	XVector axis = Axis;
	axis.Normalise();
	XReal sin_a = static_cast<XReal>(sin( Angle / 2.0f ));
	XReal cos_a = static_cast<XReal>(cos( Angle / 2.0f ));
	x    = axis.x * sin_a;
	y    = axis.y * sin_a;
	z    = axis.z * sin_a;
	w    = cos_a;
}

//--------------------------------------------------------------------------
/// \brief	This function converts a quaternion to an axis angle rotation
/// \param	Axis	-	the output axis
/// \param	Angle	-	the output angle.
///
void XQuaternion::ToAxisAngle( XVector& Axis, float &Angle )
{
	XQuaternion Q( this->Normalised() );
	float cos_a = w;
	Angle = static_cast<float>(acos( cos_a ) * 2.0f);
	float sin_a = static_cast<XReal>(sqrt( 1.0f - cos_a * cos_a ));
	if ( fabs( sin_a ) < 0.0005f )
		sin_a = 1.0f;
	Axis.x = x / sin_a;
	Axis.y = y / sin_a;
	Axis.z = z / sin_a;
}

//--------------------------------------------------------------------------
/// \brief	This function converts a set of euler angles to a quaternion
/// \param	ax	-	rotation in x
/// \param	ay	-	rotation in y
/// \param	az	-	rotation in z
///
void XQuaternion::FromEulerAngles( const float ax, const float ay, const float az )
{
	XVector vx( 1, 0, 0 ),
		vy( 0, 1, 0 ),
		vz( 0, 0, 1 );
	XQuaternion qx, qy, qz, qt;
	qx.FromAxisAngle( vx, ax );
	qy.FromAxisAngle( vy, ay );
	qz.FromAxisAngle( vz, az );
	qt = qx * qy;
	*this = qt * qz;
}


void XQuaternion::FromMatrix(const XMatrix &mat) {
	float T = 1 + mat.data[0] + mat.data[5] + mat.data[10];
	if ( T > 0.00000001f ) //to avoid large distortions!
	{
		float S = static_cast<float>(sqrt(T) * 2.0f);
		x = ( mat.data[6] - mat.data[9] ) / S;
		y = ( mat.data[8] - mat.data[2] ) / S;
		z = ( mat.data[1] - mat.data[4] ) / S;
		w = 0.25f * S;
	}
	else
	if ( mat.data[0] > mat.data[5] &&
	     mat.data[0] > mat.data[10] )  {		// Column 0:
		float S  = static_cast<float>(sqrt( 1.0f + mat.data[0] - mat.data[5] - mat.data[10] ) * 2.0f);
		x = 0.25f * S;
		y = (mat.data[1] + mat.data[4] ) / S;
		z = (mat.data[8] + mat.data[2] ) / S;
		w = (mat.data[6] - mat.data[9] ) / S;
	} 
	else 
	if ( mat.data[5] > mat.data[10] ) {			// Column 1:
		float S  = static_cast<float>(sqrt( 1.0 + mat.data[5] - mat.data[0] - mat.data[10] ) * 2.0f);
		x = (mat.data[1] + mat.data[4] ) / S;
		y = 0.25f * S;
		z = (mat.data[6] + mat.data[9] ) / S;
		w = (mat.data[8] - mat.data[2] ) / S;
	}
	else {										// Column 2:
		float S  = static_cast<float>(sqrt( 1.0 + mat.data[10] - mat.data[0] - mat.data[5] ) * 2.0f);
		x = (mat.data[8] + mat.data[2] ) / S;
		y = (mat.data[6] + mat.data[9] ) / S;
		z = 0.25f * S;
		w = (mat.data[1] - mat.data[4] ) / S;
	}
}

void XQuaternion::AsMatrix(OUT float* matrix) const
{
	// sacrafice a few bytes to pre-calculate some values
	float xx = x * x;
	float xy = x * y;
	float xz = x * z;
	float xw = x * w;
	float yy = y * y;
	float yz = y * z;
	float yw = y * w;
	float zz = z * z;
	float zw = z * w;

	// [0] -> [3]
	*(  matrix) = 1.0f - 2.0f * (yy+zz);
	*(++matrix) =        2.0f * (xy+zw);
	*(++matrix) =        2.0f * (xz-yw);
	*(++matrix) =		 0.0f;

	// [4] -> [7]
	*(++matrix) =        2.0f * (xy-zw);
	*(++matrix) = 1.0f - 2.0f * (xx+zz);
	*(++matrix) =        2.0f * (yz+xw);
	*(++matrix) =		 0.0f;

	// [8] -> [11]
	*(++matrix) =        2.0f * (xz+yw);
	*(++matrix) =        2.0f * (yz-xw);
	*(++matrix) = 1.0f - 2.0f * (xx+yy);
	*(++matrix) =		 0.0f;

	// [12] -> [15]
	*(++matrix) =		 0.0f;
	*(++matrix) =		 0.0f;
	*(++matrix) =		 0.0f;
	*(++matrix) =		 1.0f;
}


/// a function to spherically interpolate the two quaternions
//
XQuaternion SLERP(const XQuaternion &q1,const XQuaternion &q2,const float &t)
{
	// for some reason, though why I don't actually know, the second quaternion
	// needs to be inverted if the angle between them is greater than 90 begrees.
	// I think it is to control the direction of the angular rotation so that the
	// bone always slerps through the shortest angle. Just guessing though so don't
	// take my word for it....
	//
	XQuaternion Second;

	// will hold the final position
	XQuaternion Final;

	// if two quarternions are the same
	if ( FCOMPARE( q1.x, q2.x ) &&
		 FCOMPARE( q1.y, q2.y ) &&
		 FCOMPARE( q1.z, q2.z ) &&
		 FCOMPARE( q1.w, q2.w ) )
	{
		// no point in slerping them....
		return q1;
	}

	// perform dot product
	float product = ( q1.x * q2.x ) +
					( q1.y * q2.y ) +
					( q1.z * q2.z ) +
					( q1.w * q2.w );

	// if the angle is greater than 90degress, negate quaternion and product
	if ( product < 0.001f )
	{
		Second  = XQuaternion(-q2.x,-q2.y,-q2.z,-q2.w);
		product = -product;
	}
	else
	{
		Second = q2;
	}

	// calculate the linear t values for the interpolation
	float interpolating_t    = t;
	float interpolating_invt = 1.0f-t;

	if(product > 1) 
		product=0.999999f;

	if(product < -1) 
		product=-0.999999f;

	// spherical interpolation takes place around a sphere. Therefore we use the
	// product (remeber dot product to finds angle between 2 vectors) to get the
	// angle we have to rotate to go from q1 to q2.
	//
	float theta    = static_cast<float>( acos(product) );

	// precalculate the sin value
	//
	float sinTheta = static_cast<float>( sin(theta) );

	if(sinTheta != 0.0f)
	{
		interpolating_invt = static_cast<float>( sin(interpolating_invt * theta) / sinTheta   );
		interpolating_t    = static_cast<float>( sin(interpolating_t    * theta) / sinTheta   );
	}

	// perform the interpolation
	Final.x = (interpolating_invt * q1.x) + (interpolating_t * Second.x);
	Final.y = (interpolating_invt * q1.y) + (interpolating_t * Second.y);
	Final.z = (interpolating_invt * q1.z) + (interpolating_t * Second.z);
	Final.w = (interpolating_invt * q1.w) + (interpolating_t * Second.w);

	Final.Normalise();

	// return the result of interpolation
	return Final;
}

XQuaternion::XQuaternion( const XMatrix& m )
{
	// calculate trace on matrix
	float T = m.m_00 + m.m_11 + m.m_22 + 1.0f;

	// if the trace of the matrix is greater than zero perform an instant calculation.
	if( T > 0 )
	{
		float S = static_cast<float>(0.5f / sqrt(T));
		x = ( m.m_12 - m.m_21 ) * S;
		y = ( m.m_20 - m.m_02 ) * S;
		z = ( m.m_01 - m.m_10 ) * S;
		w = 0.25f / S;
	}
	else
	{
		int ind=0;
		float comp = m.m_00;
		if(comp < m.m_11)
		{
			comp = m.m_11;
			ind = 1;
		}
		if(comp < m.m_22)
		{
			comp = m.m_22;
			ind = 2;
		}
		switch(ind)
		{
		case 0:
			{
				float _1_over_S  = static_cast<float>(0.5f/sqrt( 1.0f + m.m_00 - m.m_11 - m.m_22 ));

				x = 0.5f * _1_over_S;
				y = (m.m_10 + m.m_01 ) * _1_over_S;
				z = (m.m_20 + m.m_02 ) * _1_over_S;
				w = (m.m_21 + m.m_12 ) * _1_over_S;
			}
			break;

		case 1:
			{
				float _1_over_S  = static_cast<float>(0.5f/sqrt( 1.0f - m.m_00 + m.m_11 - m.m_22 ));

				x = (m.m_10 + m.m_01 ) * _1_over_S;
				y = 0.5f * _1_over_S;
				z = (m.m_21 + m.m_12 ) * _1_over_S;
				w = (m.m_20 + m.m_02 ) * _1_over_S;
			}
			break;

		case 2:
			{
				float _1_over_S  = static_cast<float>(0.5f/sqrt( 1.0f - m.m_00 - m.m_11 + m.m_22 ));

				x = (m.m_20 + m.m_02 ) * _1_over_S;
				y = (m.m_21 + m.m_12 ) * _1_over_S;
				z = 0.5f * _1_over_S;
				w = (m.m_10 + m.m_01 ) * _1_over_S;
			}
			break;
		}

	}

}
