//=========================================================================
///
///	\file	Quaternion.h
/// \author	Rob Bateman
///	\brief	Defines the class XQuaternion, a class definition for a Quaternion
///
//=========================================================================


//=========================================================================
//	Sentry
//=========================================================================

//	This directive is used to stop the header file from being included
//	Twice and is standard for all headers.
#ifdef WIN32
#pragma once
#endif


#ifndef QUATERNION_H_INCLUDED
#define QUATERNION_H_INCLUDED

//=========================================================================
//	Includes
//=========================================================================

#include <iostream>
#include "MathDefs.h"
#include "Vector.h"
#include "Interpolation.h"

//=========================================================================
//	Class Definitions
//=========================================================================


#ifndef OUT
#define OUT
#endif
#ifndef IN
#define IN
#endif

class XMatrix;


//------------------------------------------------------------------------------
/// \brief	The XQuaternion class is used to hold rotational values for both 
///			transforms and animation keyframes. 
/// 
class XQuaternion
{
	//--------------------------------------------------------------------------
	/// \brief	this function spherically interpolates between two quaternions with respect to t
	/// \param	q1	-	the first quaternion
	/// \param	q2	-	the second quaternion
	/// \param	t	-	the interpolating t value
	/// 
	friend XQuaternion SLERP(IN const XQuaternion &q1,
							 IN const XQuaternion &q2,
							 IN const float &t);

	//--------------------------------------------------------------------------
	/// \brief	the streaming operator for the quaternion
	/// \param	ifs	-	the input stream
	/// \param	q	-	the place to read the values into 
	/// 
	friend std::istream& operator >> (IN std::istream& ifs,IN XQuaternion &q) {
		return ifs >> q.x >> q.y >> q.z >> q.w;
	}
	//--------------------------------------------------------------------------
	/// \brief	the streaming operator for the quaternion
	/// \param	ifs	-	the input stream
	/// \param	q	-	the place to read the values into 
	/// 
	friend std::ostream& operator << (IN std::ostream& ifs,IN XQuaternion &q) {
		return ifs << q.x << " " << q.y << " " << q.z << " " << q.w;
	}

public:

	//-------------------------------------------------------------------------
	//-------------------------------------------------------------------------
	//						#######		OPERATORS		#####
	//-------------------------------------------------------------------------
	//-------------------------------------------------------------------------


	//--------------------------------------------------------------------------
	/// constructor
	/// \param	x_	-	the x component of the quaternion
	/// \param	y_	-	the y component of the quaternion
	/// \param	z_	-	the z component of the quaternion 
	/// \param	w_	-	the w component of the quaternion
	/// 
	inline XQuaternion(IN XReal x_=0,
					   IN XReal y_=0,
					   IN XReal z_=0,
					   IN XReal w_=1)
		: x(x_),y(y_),z(z_),w(w_) {}

	//--------------------------------------------------------------------------
	/// copy constructor
	/// \param	q	-	the quaternion to copy
	/// 
	inline XQuaternion(IN const XQuaternion& q) 
		: x(q.x),y(q.y),z(q.z),w(q.w) {}

	


	//-------------------------------------------------------------------------
	//-------------------------------------------------------------------------
	//						#######		OPERATORS		#####
	//-------------------------------------------------------------------------
	//-------------------------------------------------------------------------

	//--------------------------------------------------------------------------
	/// \brief	returns the inverse of the quaternion (aka conjugate)
	/// \return	the conjugate of the quaternion
	/// 
	inline OUT XQuaternion operator-() const {
		return XQuaternion( -x,-y,-z,w );
	}

	//--------------------------------------------------------------------------
	/// \brief	Perform a multiplication between 2 quaternions
	///
	///			ResultQuaternion = Q1.Q2 = ( w1.w2 - v1.v2, w1.v2 + w2.v1 + v1 x v2 )
	///
	///			where   v1 = (x,y,z) of Q1
	///					w1 = (w)     of Q1
	///					v2 = (x,y,z) of Q2
	///					w2 = (w)     of Q2
	///
	/// \return	the result of the mutliplication
	///
	inline XQuaternion operator *(const XQuaternion& q) const
	{

		// create a var to hold the vector part of the two quaternions
		XVector v1( x, y, z );
		XVector v2( q.x, q.y, q.z );

		float scalar = (w-q.w) - v1 * v2;
		XVector vector_part = (w * v2) + (q.w * v1) + (v1 ^ v2);

		// return the resulting normalised quaternion
		return XQuaternion( vector_part.x, vector_part.y, vector_part.z, scalar ).Normalise();	
	}

	//--------------------------------------------------------------------------
	///	\brief	
	///	\note	A rotation matrix may be converted back to a quaternion.
	///
	XQuaternion( const XMatrix& m );

	//-------------------------------------------------------------------------
	//-------------------------------------------------------------------------
	//					#######		 FUNCTIONS		#####
	//-------------------------------------------------------------------------
	//-------------------------------------------------------------------------

	//--------------------------------------------------------------------------
	/// \brief	This function normalises the quaternion
	/// \return	A constant reference to itself
	/// 
	inline const XQuaternion& Normalise() {
		XReal l_inv = 1.0f/Mag();
		x*=l_inv;
		y*=l_inv;
		z*=l_inv;
		w*=l_inv;
		return (*this);
	}

	//--------------------------------------------------------------------------
	/// \brief	returns a normalised version of this quaternion
	/// \return	The normalised quaternion 
	/// 
	inline OUT XQuaternion Normalised() const {
		float l_inv = 1.0f/Mag();
		return XQuaternion (x*l_inv,y*l_inv,z*l_inv,w*l_inv);
	}

	//--------------------------------------------------------------------------
	/// \brief	this function turns the quaternion into a 4x4 rotation matrix
	/// \param	matrix	-	the output matrix
	///
	void AsMatrix(OUT XReal* matrix) const;

	//--------------------------------------------------------------------------
	/// \brief	This function returns the quaternion as a matrix
	/// \return	The rotation matrix equivalent
	///
	OUT XMatrix AsMatrix() const;

	//--------------------------------------------------------------------------
	/// \brief	This function extracts a quaternion from the rotations
	/// 		present in a matrix.
	/// \param	mat	-	the input matrix
	///
	void FromMatrix(const XMatrix &mat);


	//--------------------------------------------------------------------------
	/// \brief	This function converts an axis angle rotation to a quaternion
	/// \param	Axis	-	the axis around which the rotation occurs
	/// \param	Angle	-	the amount of rotation
	///
	void FromAxisAngle(const XVector& Axis,const float Angle) ;

	//--------------------------------------------------------------------------
	/// \brief	This function converts a quaternion to an axis angle rotation
	/// \param	Axis	-	the output axis
	/// \param	Angle	-	the output angle.
	///
	void ToAxisAngle( XVector& Axis, float &Angle );
	
	//--------------------------------------------------------------------------
	/// \brief	This function converts a set of euler angles to a quaternion
	/// \param	ax	-	rotation in x
	/// \param	ay	-	rotation in y
	/// \param	az	-	rotation in z
	///
	void FromEulerAngles( const float ax, const float ay, const float az );


	//--------------------------------------------------------------------------
	/// \brief	returns the magnitude of the quaternion
	/// \return	The magnitude of the quaternion 
	/// 
	inline OUT XReal Mag() const {
		return static_cast<XReal>( sqrt(w*w + x*x + y*y + z*z) );
	}

	/// the quaternion data
	XReal x;
	XReal y;
	XReal z;
	XReal w;
};

#ifndef EPISILON
#define EPISILON 0.001f
#endif

static bool DIFF(XQuaternion& a,XQuaternion& b) {
	XQuaternion c(  b.x-a.x,
			b.y-a.y,
			b.z-a.z,
			b.w-a.w );

	if( fabs(c.w) > EPISILON ||
	    fabs(c.x) > EPISILON ||
	    fabs(c.y) > EPISILON ||
	    fabs(c.z) > EPISILON ) {
	    return true;
	}
	return false;
}

static bool DIFF(XVector& a,XVector& b) {
	XVector c = b-a;

	if( fabs(c.x) > EPISILON ||
	    fabs(c.y) > EPISILON ||
	    fabs(c.z) > EPISILON ) {
	    return true;
	}
	return false;
}

template<>
static void LERP<XQuaternion>(OUT XQuaternion& result,IN const float t,IN const XQuaternion& v0,IN const XQuaternion& v1) {
	result = SLERP(v0,v1,t);
}
template<>
static XQuaternion LERP<XQuaternion>(IN const float t,IN const XQuaternion& v0,IN const XQuaternion& v1) {
	return SLERP(v0,v1,t);
}


#include "Matrix.h"

#endif
