//=========================================================================
///
///	\file	Sphere.h
/// \author	Rob
///	\brief	Defines the class XSphere which defines a 3D sphere withe a center
///			point and a radius. 
///
//=========================================================================


//=========================================================================
//	Sentry
//=========================================================================
#ifdef WIN32
#pragma once
#endif


#ifndef SPHERE_H_INCLUDED
#define SPHERE_H_INCLUDED

//=========================================================================
//	Includes
//=========================================================================

#include "MathDefs.h"
#include <iostream>
#include "Vector.h"
#include "Line.h"



//=========================================================================
//	Class Definitions
//=========================================================================

class XSphere
{
public :

	/// \brief	This function is used to determine if the specified point p
	///			is within the bounds of the sphere
	/// \param	p	-	the point to test
	/// \return	true if in sphere, false otherwise
	/// 
	inline bool PointInSphere(IN const XVector& p) const 
	{
		return (Center-p).Length() < Radius;
	}

	/// \brief	This function is used to see if the specified line intersects the
	///			sphere. 
	/// \param	line	-	The line to test
	/// \return true if it intersects, false otherwise
	/// 
	inline bool LineIntersects(IN const XLine& line) const
	{
		// find the closest point to the spheres center along the line
		XVector testP = line.ClosestPoint(Center);

		// test to see if that point lies in the sphere 
		return PointInSphere(testP);
	}

	/// \brief	This function is used to see if the line created from the two points 
	///			specified intersects the sphere. 
	/// \param	p1	-	The first point of the line to test
	/// \param	p2	-	The second point of the line to test
	/// \return true if it intersects, false otherwise
	/// 
	inline bool LineIntersects(IN const XVector& p1,IN const XVector& p2) const
	{
		return LineIntersects( XLine(p1,p2) );
	}

	/// \brief	This function is used to get hold of the line intersection points when a line
	///			intersects a sphere. There are 3 possibilities, either 
	///				1. the line does not intersect the sphere
	///				2. The line intersects the sphere once (one point is inside the sphere)
	///				3. The line intersects the sphere twice
	/// \param	p1		-	This is used to store the first intersect point. If the line does not 
	///						intersect the sphere then ignore this value
	/// \param	p2		-	This is used to store the second intersect point. If the line does not 
	///						intersect the sphere, or only one intersect exists then ignore this value
	/// \param	line	-	the line to test for intersection with this sphere
	/// \return 0 if no intersections exists
	///			1 if one intersection point
	///			2 if 2 intersections.
	/// 
	unsigned int GetLineIntersects(OUT XVector& p1,OUT XVector& p2,IN const XLine& line) const;

	/// \brief	This function takes a specified point and returns the nearest point
	///			on the spheres surface
	/// \param	p	-	the point to test
	/// \return	the nearest point on the surface of the sphere
	/// 
	inline XVector ClosestPoint(const XVector& p) const 
	{
		// get the direction from the center of the sphere to the point being tested.
		//
		XVector dir = (p-Center).Normalise();

		// closest point is found by projecting down the direction from the center 
		// point of the sphere by the radius....
		// 
		return Center + dir * Radius;
	}



	/// the center of the sphere
	XVector Center;

	/// the radius of the sphere
	XReal	Radius;
};

#endif

